"""Planar intrinsics calibration from real stereo dataset images.

Python counterpart of:
`crates/vision-calibration/examples/planar_real.rs`
"""

from __future__ import annotations

import re
from pathlib import Path

import calib_targets
import numpy as np
from PIL import Image
import vision_calibration as vc

from _common import repo_root

BOARD_ROWS = 7
BOARD_COLS = 11
SQUARE_SIZE_M = 0.03


def load_gray(path: Path) -> np.ndarray:
    return np.asarray(Image.open(path).convert("L"), dtype=np.uint8)


def image_indices(folder: Path, prefix: str) -> list[int]:
    pattern = re.compile(rf"^{re.escape(prefix)}(\d+)\.png$")
    out: list[int] = []
    for path in folder.glob("*.png"):
        m = pattern.match(path.name)
        if m is not None:
            out.append(int(m.group(1)))
    out.sort()
    return out


def detect_chessboard_observation(path: Path) -> vc.Observation | None:
    image = load_gray(path)
    detection = calib_targets.detect_chessboard(
        image,
        chess_cfg=calib_targets.ChessConfig(),
        params=calib_targets.ChessboardParams(
            expected_rows=BOARD_ROWS,
            expected_cols=BOARD_COLS,
        ),
    )
    if detection is None:
        return None

    points_3d: list[tuple[float, float, float]] = []
    points_2d: list[tuple[float, float]] = []
    for corner in detection.detection.corners:
        if corner.grid is None:
            continue
        points_3d.append((float(corner.grid.i) * SQUARE_SIZE_M, float(corner.grid.j) * SQUARE_SIZE_M, 0.0))
        points_2d.append((float(corner.position[0]), float(corner.position[1])))
    if len(points_2d) < 4:
        return None
    return vc.Observation(points_3d=points_3d, points_2d=points_2d)


def load_dataset() -> vc.PlanarDataset:
    left_dir = repo_root() / "data/stereo/imgs/leftcamera"
    if not left_dir.exists():
        raise FileNotFoundError(f"dataset not found: {left_dir}")

    views: list[vc.PlanarView] = []
    skipped = 0
    for idx in image_indices(left_dir, "Im_L_"):
        obs = detect_chessboard_observation(left_dir / f"Im_L_{idx}.png")
        if obs is None:
            skipped += 1
            continue
        views.append(vc.PlanarView(observation=obs))

    if len(views) < 3:
        raise RuntimeError(f"need at least 3 usable views, got {len(views)}")
    print(f"Loaded {len(views)} views ({skipped} skipped)")
    return vc.PlanarDataset(views=views)


def main() -> None:
    print("=== Planar Intrinsics Calibration (Real Images, Python) ===")
    dataset = load_dataset()
    result = vc.run_planar_intrinsics(dataset)
    k = result.camera.intrinsics
    d = result.camera.distortion

    print(f"Final cost: {result.final_cost:.3e}")
    print(f"Mean reprojection error: {result.mean_reproj_error:.6f} px")
    print(f"Intrinsics: fx={k.fx:.3f}, fy={k.fy:.3f}, cx={k.cx:.3f}, cy={k.cy:.3f}")
    print(f"Distortion: k1={d.k1:.6f}, k2={d.k2:.6f}, k3={d.k3:.6f}, p1={d.p1:.6f}, p2={d.p2:.6f}")


if __name__ == "__main__":
    main()



"""Python bindings for calibration-rs.

High-level workflow functions accept Python-native dataclasses from
``vision_calibration.models`` and return result dataclasses with docstrings.
Raw serde payload access is available only via low-level helpers in
``vision_calibration._api``.
"""

from __future__ import annotations

from . import _vision_calibration as _native
from ._api import (
    robust_arctan,
    robust_cauchy,
    robust_huber,
    robust_none,
    run_laserline_device,
    run_planar_intrinsics,
    run_rig_extrinsics,
    run_rig_handeye,
    run_scheimpflug_intrinsics,
    run_single_cam_handeye,
)
from .models import (
    BrownConradyDistortion,
    LaserlineDataset,
    LaserlineDeviceCalibrationConfig,
    LaserlineEstimate,
    LaserlineEstimateParams,
    LaserlineDeviceInitConfig,
    LaserlineDeviceOptimizeConfig,
    LaserlineDeviceResult,
    LaserlineDeviceSolverConfig,
    LaserlinePlane,
    LaserlineStats,
    LaserlineView,
    Observation,
    PinholeBrownConradyCamera,
    PinholeBrownConradyScheimpflugCamera,
    PinholeIntrinsics,
    PlanarCalibrationConfig,
    PlanarCalibrationResult,
    PlanarDataset,
    PlanarView,
    Pose,
    ScheimpflugIntrinsicsCalibrationConfig,
    ScheimpflugSensor,
    ScheimpflugIntrinsicsResult,
    RigExtrinsicsCalibrationConfig,
    RigExtrinsicsDataset,
    RigExtrinsicsResult,
    RigExtrinsicsView,
    RigHandeyeBaConfig,
    RigHandeyeCalibrationConfig,
    RigHandeyeDataset,
    RigHandeyeInitConfig,
    RigHandeyeIntrinsicsConfig,
    RigHandeyeResult,
    RigHandeyeRigConfig,
    RigHandeyeSolverConfig,
    RigHandeyeView,
    SingleCamHandeyeCalibrationConfig,
    SingleCamHandeyeDataset,
    SingleCamHandeyeResult,
    SingleCamHandeyeView,
)
__version__ = _native.library_version()

__all__ = [
    "__version__",
    "Pose",
    "Observation",
    "PlanarView",
    "PlanarDataset",
    "SingleCamHandeyeView",
    "SingleCamHandeyeDataset",
    "RigExtrinsicsView",
    "RigExtrinsicsDataset",
    "RigHandeyeView",
    "RigHandeyeDataset",
    "LaserlineView",
    "LaserlineDataset",
    "PinholeIntrinsics",
    "BrownConradyDistortion",
    "ScheimpflugSensor",
    "PinholeBrownConradyCamera",
    "PinholeBrownConradyScheimpflugCamera",
    "LaserlinePlane",
    "LaserlineEstimateParams",
    "LaserlineEstimate",
    "LaserlineStats",
    "PlanarCalibrationConfig",
    "ScheimpflugIntrinsicsCalibrationConfig",
    "SingleCamHandeyeCalibrationConfig",
    "RigExtrinsicsCalibrationConfig",
    "RigHandeyeIntrinsicsConfig",
    "RigHandeyeRigConfig",
    "RigHandeyeInitConfig",
    "RigHandeyeSolverConfig",
    "RigHandeyeBaConfig",
    "RigHandeyeCalibrationConfig",
    "LaserlineDeviceInitConfig",
    "LaserlineDeviceSolverConfig",
    "LaserlineDeviceOptimizeConfig",
    "LaserlineDeviceCalibrationConfig",
    "PlanarCalibrationResult",
    "ScheimpflugIntrinsicsResult",
    "SingleCamHandeyeResult",
    "RigExtrinsicsResult",
    "RigHandeyeResult",
    "LaserlineDeviceResult",
    "robust_none",
    "robust_huber",
    "robust_cauchy",
    "robust_arctan",
    "run_planar_intrinsics",
    "run_scheimpflug_intrinsics",
    "run_single_cam_handeye",
    "run_rig_extrinsics",
    "run_rig_handeye",
    "run_laserline_device",
]



"""Stereo rig extrinsics calibration from real stereo images.

Python counterpart of:
`crates/vision-calibration/examples/stereo_session.rs`
"""

from __future__ import annotations

import re
from pathlib import Path

import calib_targets
import numpy as np
from PIL import Image
import vision_calibration as vc

from _common import baseline_m, repo_root

BOARD_ROWS = 7
BOARD_COLS = 11
SQUARE_SIZE_M = 0.03


def load_gray(path: Path) -> np.ndarray:
    return np.asarray(Image.open(path).convert("L"), dtype=np.uint8)


def list_pair_indices(left_dir: Path, right_dir: Path) -> list[int]:
    left_pat = re.compile(r"^Im_L_(\d+)\.png$")
    right_pat = re.compile(r"^Im_R_(\d+)\.png$")
    left = set()
    right = set()
    for path in left_dir.glob("*.png"):
        m = left_pat.match(path.name)
        if m:
            left.add(int(m.group(1)))
    for path in right_dir.glob("*.png"):
        m = right_pat.match(path.name)
        if m:
            right.add(int(m.group(1)))
    return sorted(left.intersection(right))


def detect_chessboard_observation(path: Path) -> vc.Observation | None:
    image = load_gray(path)
    detection = calib_targets.detect_chessboard(
        image,
        chess_cfg=calib_targets.ChessConfig(),
        params=calib_targets.ChessboardParams(
            expected_rows=BOARD_ROWS,
            expected_cols=BOARD_COLS,
        ),
    )
    if detection is None:
        return None

    points_3d: list[tuple[float, float, float]] = []
    points_2d: list[tuple[float, float]] = []
    for corner in detection.detection.corners:
        if corner.grid is None:
            continue
        points_3d.append((float(corner.grid.i) * SQUARE_SIZE_M, float(corner.grid.j) * SQUARE_SIZE_M, 0.0))
        points_2d.append((float(corner.position[0]), float(corner.position[1])))
    if len(points_2d) < 4:
        return None
    return vc.Observation(points_3d=points_3d, points_2d=points_2d)


def load_dataset() -> vc.RigExtrinsicsDataset:
    base = repo_root() / "data/stereo/imgs"
    left_dir = base / "leftcamera"
    right_dir = base / "rightcamera"
    if not left_dir.exists() or not right_dir.exists():
        raise FileNotFoundError(f"stereo dataset not found: {base}")

    views: list[vc.RigExtrinsicsView] = []
    skipped = 0
    left_ok = 0
    right_ok = 0
    for idx in list_pair_indices(left_dir, right_dir):
        left = detect_chessboard_observation(left_dir / f"Im_L_{idx}.png")
        right = detect_chessboard_observation(right_dir / f"Im_R_{idx}.png")
        if left is None and right is None:
            skipped += 1
            continue
        left_ok += int(left is not None)
        right_ok += int(right is not None)
        views.append(vc.RigExtrinsicsView(cameras=[left, right]))

    if len(views) < 3:
        raise RuntimeError(f"need at least 3 usable views, got {len(views)}")
    print(f"Loaded {len(views)} views ({skipped} skipped), usable: left={left_ok}, right={right_ok}")
    return vc.RigExtrinsicsDataset(num_cameras=2, views=views)


def main() -> None:
    print("=== Stereo Rig Calibration Session (Real Images, Python) ===")
    dataset = load_dataset()
    result = vc.run_rig_extrinsics(dataset)
    base_mm = baseline_m(result.cam_se3_rig) * 1000.0

    print(f"Mean reprojection error: {result.mean_reproj_error:.6f} px")
    print(f"Per-camera reprojection: {result.per_cam_reproj_errors}")
    print(f"Estimated baseline: {base_mm:.3f} mm")
    for idx, cam in enumerate(result.cameras):
        k = cam.intrinsics
        d = cam.distortion
        print(f"Camera {idx}: fx={k.fx:.3f}, fy={k.fy:.3f}, cx={k.cx:.3f}, cy={k.cy:.3f}")
        print(f"  Distortion: k1={d.k1:.6f}, k2={d.k2:.6f}, p1={d.p1:.6f}, p2={d.p2:.6f}")


if __name__ == "__main__":
    main()



"""High-level Python wrappers around the Rust extension.

Public runner functions accept typed dataclasses from :mod:`vision_calibration.models`
and return typed result dataclasses.

Low-level raw helpers (prefixed with ``_run_*_raw``) are kept for interop paths
that need direct serde payload control.
"""

from __future__ import annotations

from typing import Any, Mapping, TypeVar, cast

from . import _vision_calibration as _native
from .models import (
    LaserlineDataset,
    LaserlineDeviceCalibrationConfig,
    LaserlineDeviceResult,
    PlanarCalibrationConfig,
    PlanarCalibrationResult,
    PlanarDataset,
    RigExtrinsicsCalibrationConfig,
    RigExtrinsicsDataset,
    RigExtrinsicsResult,
    RigHandeyeCalibrationConfig,
    RigHandeyeDataset,
    RigHandeyeResult,
    ScheimpflugIntrinsicsCalibrationConfig,
    ScheimpflugIntrinsicsResult,
    SingleCamHandeyeCalibrationConfig,
    SingleCamHandeyeDataset,
    SingleCamHandeyeResult,
)
from .types import RobustLoss

_T = TypeVar("_T")


def robust_none() -> RobustLoss:
    """Build a serde-compatible "no robust loss" value."""
    return "None"


def robust_huber(scale: float) -> RobustLoss:
    """Build a serde-compatible Huber robust loss payload."""
    return {"Huber": {"scale": float(scale)}}


def robust_cauchy(scale: float) -> RobustLoss:
    """Build a serde-compatible Cauchy robust loss payload."""
    return {"Cauchy": {"scale": float(scale)}}


def robust_arctan(scale: float) -> RobustLoss:
    """Build a serde-compatible Arctan robust loss payload."""
    return {"Arctan": {"scale": float(scale)}}


def _ensure_type(value: Any, expected_type: type[_T], name: str) -> _T:
    if not isinstance(value, expected_type):
        raise TypeError(
            f"{name} must be {expected_type.__name__}, got {type(value).__name__}"
        )
    return value


def _ensure_optional_type(
    value: Any,
    expected_type: type[_T],
    name: str,
) -> _T | None:
    if value is None:
        return None
    return _ensure_type(value, expected_type, name)


# ─────────────────────────────────────────────────────────────────────────────
# Low-level raw helpers
# ─────────────────────────────────────────────────────────────────────────────


def _run_planar_intrinsics_raw(
    input_payload: Mapping[str, Any],
    config_payload: Mapping[str, Any] | None = None,
) -> PlanarCalibrationResult:
    raw = cast(dict[str, Any], _native.run_planar_intrinsics(input_payload, config_payload))
    return PlanarCalibrationResult.from_payload(raw)


def _run_single_cam_handeye_raw(
    input_payload: Mapping[str, Any],
    config_payload: Mapping[str, Any] | None = None,
) -> SingleCamHandeyeResult:
    raw = cast(dict[str, Any], _native.run_single_cam_handeye(input_payload, config_payload))
    return SingleCamHandeyeResult.from_payload(raw)


def _run_rig_extrinsics_raw(
    input_payload: Mapping[str, Any],
    config_payload: Mapping[str, Any] | None = None,
) -> RigExtrinsicsResult:
    raw = cast(dict[str, Any], _native.run_rig_extrinsics(input_payload, config_payload))
    return RigExtrinsicsResult.from_payload(raw)


def _run_rig_handeye_raw(
    input_payload: Mapping[str, Any],
    config_payload: Mapping[str, Any] | None = None,
) -> RigHandeyeResult:
    raw = cast(dict[str, Any], _native.run_rig_handeye(input_payload, config_payload))
    return RigHandeyeResult.from_payload(raw)


def _run_laserline_device_raw(
    input_payload: list[dict[str, Any]],
    config_payload: Mapping[str, Any] | None = None,
) -> LaserlineDeviceResult:
    raw = cast(dict[str, Any], _native.run_laserline_device(input_payload, config_payload))
    return LaserlineDeviceResult.from_payload(raw)


def _run_scheimpflug_intrinsics_raw(
    input_payload: Mapping[str, Any],
    config_payload: Mapping[str, Any] | None = None,
) -> ScheimpflugIntrinsicsResult:
    raw = cast(
        dict[str, Any],
        _native.run_scheimpflug_intrinsics(input_payload, config_payload),
    )
    return ScheimpflugIntrinsicsResult.from_payload(raw)


# ─────────────────────────────────────────────────────────────────────────────
# High-level typed API
# ─────────────────────────────────────────────────────────────────────────────


def run_planar_intrinsics(
    input: PlanarDataset,
    config: PlanarCalibrationConfig | None = None,
) -> PlanarCalibrationResult:
    """Run planar intrinsics calibration with typed input/config objects."""
    dataset = _ensure_type(input, PlanarDataset, "input")
    cfg = _ensure_optional_type(config, PlanarCalibrationConfig, "config")
    return _run_planar_intrinsics_raw(
        dataset.to_payload(),
        None if cfg is None else cfg.to_payload(),
    )


def run_single_cam_handeye(
    input: SingleCamHandeyeDataset,
    config: SingleCamHandeyeCalibrationConfig | None = None,
) -> SingleCamHandeyeResult:
    """Run single-camera hand-eye calibration with typed input/config objects."""
    dataset = _ensure_type(input, SingleCamHandeyeDataset, "input")
    cfg = _ensure_optional_type(config, SingleCamHandeyeCalibrationConfig, "config")
    return _run_single_cam_handeye_raw(
        dataset.to_payload(),
        None if cfg is None else cfg.to_payload(),
    )


def run_rig_extrinsics(
    input: RigExtrinsicsDataset,
    config: RigExtrinsicsCalibrationConfig | None = None,
) -> RigExtrinsicsResult:
    """Run rig extrinsics calibration with typed input/config objects."""
    dataset = _ensure_type(input, RigExtrinsicsDataset, "input")
    cfg = _ensure_optional_type(config, RigExtrinsicsCalibrationConfig, "config")
    return _run_rig_extrinsics_raw(
        dataset.to_payload(),
        None if cfg is None else cfg.to_payload(),
    )


def run_rig_handeye(
    input: RigHandeyeDataset,
    config: RigHandeyeCalibrationConfig | None = None,
) -> RigHandeyeResult:
    """Run rig hand-eye calibration with typed input/config objects."""
    dataset = _ensure_type(input, RigHandeyeDataset, "input")
    cfg = _ensure_optional_type(config, RigHandeyeCalibrationConfig, "config")
    return _run_rig_handeye_raw(
        dataset.to_payload(),
        None if cfg is None else cfg.to_payload(),
    )


def run_laserline_device(
    input: LaserlineDataset,
    config: LaserlineDeviceCalibrationConfig | None = None,
) -> LaserlineDeviceResult:
    """Run laserline-device calibration with typed input/config objects."""
    dataset = _ensure_type(input, LaserlineDataset, "input")
    cfg = _ensure_optional_type(config, LaserlineDeviceCalibrationConfig, "config")
    return _run_laserline_device_raw(
        dataset.to_payload(),
        None if cfg is None else cfg.to_payload(),
    )


def run_scheimpflug_intrinsics(
    input: PlanarDataset,
    config: ScheimpflugIntrinsicsCalibrationConfig | None = None,
) -> ScheimpflugIntrinsicsResult:
    """Run Scheimpflug intrinsics calibration with typed input/config objects."""
    dataset = _ensure_type(input, PlanarDataset, "input")
    cfg = _ensure_optional_type(config, ScheimpflugIntrinsicsCalibrationConfig, "config")
    return _run_scheimpflug_intrinsics_raw(
        dataset.to_payload(),
        None if cfg is None else cfg.to_payload(),
    )



"""Single-camera hand-eye calibration with synthetic data.

Python counterpart of:
`crates/vision-calibration/examples/handeye_synthetic.rs`
"""

from __future__ import annotations

import vision_calibration as vc

from _common import (
    grid_points,
    make_pose,
    pose_compose,
    pose_inverse,
    project_point_pinhole,
    transform_point,
)


def make_input() -> vc.SingleCamHandeyeDataset:
    intrinsics_gt = {"fx": 800.0, "fy": 780.0, "cx": 640.0, "cy": 360.0, "skew": 0.0}
    distortion_gt = {"k1": 0.0, "k2": 0.0, "k3": 0.0, "p1": 0.0, "p2": 0.0, "iters": 8}
    handeye_gt = make_pose(0.05, -0.03, 0.10, 0.10, -0.05, 0.02)  # T_G_C
    target_in_base_gt = make_pose(0.0, 0.0, 0.0, 0.0, 0.0, 1.0)  # T_B_T
    board = grid_points(cols=6, rows=5, spacing_m=0.05)

    robot_poses = [
        make_pose(0.0, 0.0, 0.0, 0.00, 0.00, 0.00),
        make_pose(0.3, 0.0, 0.0, 0.10, 0.00, 0.00),
        make_pose(0.0, 0.3, 0.0, 0.00, 0.10, 0.00),
        make_pose(0.0, 0.0, 0.3, 0.00, 0.00, 0.10),
        make_pose(0.2, 0.2, 0.0, 0.05, -0.05, 0.00),
        make_pose(-0.2, 0.0, 0.2, -0.05, 0.05, 0.00),
        make_pose(0.15, -0.15, 0.1, 0.00, -0.05, 0.05),
    ]

    views: list[vc.SingleCamHandeyeView] = []
    for robot_pose in robot_poses:
        # Eye-in-hand chain: T_C_T = (T_B_G * T_G_C)^-1 * T_B_T
        cam_pose = pose_compose(pose_inverse(pose_compose(robot_pose, handeye_gt)), target_in_base_gt)
        points_3d = []
        points_2d = []
        for p in board:
            p_cam = transform_point(cam_pose, p)
            uv = project_point_pinhole(p_cam, intrinsics_gt, distortion_gt)
            if uv is None:
                continue
            points_3d.append(p)
            points_2d.append(uv)
        if len(points_2d) >= 4:
            views.append(
                vc.SingleCamHandeyeView(
                    observation=vc.Observation(points_3d=points_3d, points_2d=points_2d),
                    base_se3_gripper=vc.Pose(
                        rotation_xyzw=tuple(robot_pose["rotation"]),
                        translation_xyz=tuple(robot_pose["translation"]),
                    ),
                )
            )
    return vc.SingleCamHandeyeDataset(views=views)


def main() -> None:
    print("=== Single-Camera Hand-Eye Calibration (Synthetic, Python) ===")
    payload = make_input()
    result = vc.run_single_cam_handeye(payload)
    k = result.camera.intrinsics

    print(f"Views: {len(payload.views)}")
    print(f"Hand-eye mode: {result.handeye_mode}")
    print(f"Mean reprojection error: {result.mean_reproj_error:.6f} px")
    print(f"Per-camera reprojection: {result.per_cam_reproj_errors}")
    print(f"Recovered intrinsics: fx={k.fx:.3f}, fy={k.fy:.3f}, cx={k.cx:.3f}, cy={k.cy:.3f}")
    if result.gripper_se3_camera is not None:
        t = result.gripper_se3_camera.translation_xyz
        print(f"Recovered gripper->camera translation: [{t[0]:.4f}, {t[1]:.4f}, {t[2]:.4f}] m")


if __name__ == "__main__":
    main()



"""Single laserline-device calibration with synthetic data.

Python counterpart of:
`crates/vision-calibration/examples/laserline_device_session.rs`
"""

from __future__ import annotations

import vision_calibration as vc

from _common import (
    add3,
    dot3,
    grid_points,
    make_pose,
    mat3_mul_vec,
    mat3_transpose,
    norm3,
    normalize3,
    pose_to_rt,
    project_point_pinhole,
    scale3,
    transform_point,
)


def make_poses() -> list[dict]:
    return [
        make_pose(0.0, 0.0, 0.0, 0.00, 0.00, 0.50),
        make_pose(0.15, -0.05, 0.0, 0.05, -0.02, 0.55),
        make_pose(-0.10, 0.08, 0.02, -0.04, 0.03, 0.60),
        make_pose(0.05, 0.12, -0.04, 0.02, 0.06, 0.52),
        make_pose(-0.12, -0.04, 0.06, -0.06, -0.04, 0.58),
    ]


def make_laser_plane(poses: list[dict]) -> dict:
    normal = normalize3([0.1, 0.05, 0.99])
    target_center = [0.075, 0.06, 0.0]
    center_cam = transform_point(poses[0], target_center)
    return {"normal": normal, "distance": -dot3(normal, center_cam)}


def laser_pixels_for_view(
    pose: dict,
    plane: dict,
    intrinsics: dict,
    distortion: dict,
) -> list[list[float]]:
    r, t = pose_to_rt(pose)
    rt = mat3_transpose(r)
    n_c = list(plane["normal"])
    n_t = mat3_mul_vec(rt, n_c)
    d_t = dot3(n_c, t) + float(plane["distance"])

    direction = [n_t[1], -n_t[0], 0.0]
    direction_norm = norm3(direction)
    if direction_norm < 1e-12:
        return []
    direction = scale3(direction, 1.0 / direction_norm)

    if abs(n_t[0]) > abs(n_t[1]):
        x0, y0 = -d_t / n_t[0], 0.0
    else:
        x0, y0 = 0.0, -d_t / n_t[1]

    pixels: list[list[float]] = []
    for i in range(40):
        s = (i / 39.0) * 0.2 - 0.1
        p_target = [x0 + s * direction[0], y0 + s * direction[1], 0.0]
        p_cam = transform_point(pose, p_target)
        uv = project_point_pinhole(p_cam, intrinsics, distortion)
        if uv is None:
            continue
        if not (uv[0] == uv[0] and uv[1] == uv[1]):
            continue
        if p_cam[2] <= 1e-6:
            continue
        ray_dir_cam = normalize3(p_cam)
        ray_dir_target = mat3_mul_vec(rt, ray_dir_cam)
        if abs(ray_dir_target[2]) <= 1e-3:
            continue
        if not (0.0 <= uv[0] < 1280.0 and 0.0 <= uv[1] < 720.0):
            continue
        pixels.append(uv)
    return pixels


def make_input() -> vc.LaserlineDataset:
    intrinsics = {"fx": 800.0, "fy": 780.0, "cx": 640.0, "cy": 360.0, "skew": 0.0}
    distortion = {"k1": 0.0, "k2": 0.0, "k3": 0.0, "p1": 0.0, "p2": 0.0, "iters": 8}
    points_3d = grid_points(cols=6, rows=5, spacing_m=0.03)
    poses = make_poses()
    plane = make_laser_plane(poses)

    views: list[vc.LaserlineView] = []
    for pose in poses:
        obs_3d: list[list[float]] = []
        obs_2d: list[list[float]] = []
        for p in points_3d:
            p_cam = transform_point(pose, p)
            uv = project_point_pinhole(p_cam, intrinsics, distortion)
            if uv is None:
                continue
            obs_3d.append(p)
            obs_2d.append(uv)

        laser_pixels = laser_pixels_for_view(pose, plane, intrinsics, distortion)
        if len(obs_2d) >= 4 and laser_pixels:
            views.append(
                vc.LaserlineView(
                    observation=vc.Observation(points_3d=obs_3d, points_2d=obs_2d),
                    laser_pixels=[tuple(p) for p in laser_pixels],
                    laser_weights=None,
                )
            )

    return vc.LaserlineDataset(views=views)


def main() -> None:
    print("=== Laserline Device Calibration (Synthetic, Python) ===")
    payload = make_input()
    result = vc.run_laserline_device(payload)
    intr = result.estimate.params.intrinsics
    plane = result.estimate.params.plane

    print(f"Views: {len(payload.views)}")
    print(f"Final cost: {result.estimate.final_cost:.3e}")
    print(f"Mean reprojection error: {result.mean_reproj_error:.6f} px")
    print(f"Mean laser error: {result.mean_laser_error:.6f}")
    print(f"Recovered intrinsics: fx={intr.fx:.3f}, fy={intr.fy:.3f}, cx={intr.cx:.3f}, cy={intr.cy:.3f}")
    print(f"Recovered plane: normal={plane.normal_xyz}, distance={plane.distance:.6f}")


if __name__ == "__main__":
    main()



"""Single-camera hand-eye calibration with real KUKA dataset images.

Python counterpart of:
`crates/vision-calibration/examples/handeye_session.rs`
"""

from __future__ import annotations

from pathlib import Path

import calib_targets
import numpy as np
from PIL import Image
import vision_calibration as vc

from _common import quat_xyzw_from_rot, repo_root

BOARD_ROWS = 17
BOARD_COLS = 28


def load_gray(path: Path) -> np.ndarray:
    return np.asarray(Image.open(path).convert("L"), dtype=np.uint8)


def parse_square_size_m(path: Path) -> float:
    raw = path.read_text(encoding="utf-8").strip().lower()
    if raw.endswith("mm"):
        return float(raw[:-2].strip()) / 1000.0
    if raw.endswith("m"):
        return float(raw[:-1].strip())
    value = float(raw)
    return value / 1000.0 if value > 1.0 else value


def parse_robot_poses(path: Path) -> list[vc.Pose]:
    poses: list[vc.Pose] = []
    for line in path.read_text(encoding="utf-8").splitlines():
        line = line.strip()
        if not line:
            continue
        values = [float(v) for v in line.split()]
        if len(values) != 16:
            raise ValueError(f"expected 16 values per pose line, got {len(values)}")
        r = [
            [values[0], values[1], values[2]],
            [values[4], values[5], values[6]],
            [values[8], values[9], values[10]],
        ]
        t = (values[3], values[7], values[11])
        poses.append(vc.Pose(rotation_xyzw=quat_xyzw_from_rot(r), translation_xyz=t))
    return poses


def detect_observation(path: Path, square_size_m: float) -> vc.Observation | None:
    image = load_gray(path)
    detection = calib_targets.detect_chessboard(
        image,
        chess_cfg=calib_targets.ChessConfig(),
        params=calib_targets.ChessboardParams(
            expected_rows=BOARD_ROWS,
            expected_cols=BOARD_COLS,
        ),
    )
    if detection is None:
        return None

    points_3d: list[tuple[float, float, float]] = []
    points_2d: list[tuple[float, float]] = []
    for corner in detection.detection.corners:
        if corner.grid is None:
            continue
        points_3d.append((float(corner.grid.i) * square_size_m, float(corner.grid.j) * square_size_m, 0.0))
        points_2d.append((float(corner.position[0]), float(corner.position[1])))
    if len(points_2d) < 4:
        return None
    return vc.Observation(points_3d=points_3d, points_2d=points_2d)


def load_dataset() -> vc.SingleCamHandeyeDataset:
    data_dir = repo_root() / "data/kuka_1"
    if not data_dir.exists():
        raise FileNotFoundError(f"dataset not found: {data_dir}")

    square_size_m = parse_square_size_m(data_dir / "squaresize.txt")
    robot_poses = parse_robot_poses(data_dir / "RobotPosesVec.txt")
    views: list[vc.SingleCamHandeyeView] = []
    skipped = 0

    for idx, robot_pose in enumerate(robot_poses, start=1):
        img_path = data_dir / f"{idx:02}.png"
        if not img_path.exists():
            skipped += 1
            continue
        obs = detect_observation(img_path, square_size_m)
        if obs is None:
            skipped += 1
            continue
        views.append(vc.SingleCamHandeyeView(observation=obs, base_se3_gripper=robot_pose))

    if len(views) < 3:
        raise RuntimeError(f"need at least 3 usable views, got {len(views)}")
    print(f"Loaded {len(views)} views ({skipped} skipped)")
    return vc.SingleCamHandeyeDataset(views=views)


def main() -> None:
    print("=== Single-Camera Hand-Eye Session (Real Images, Python) ===")
    dataset = load_dataset()
    result = vc.run_single_cam_handeye(dataset)
    cam = result.camera.intrinsics

    print(f"Hand-eye mode: {result.handeye_mode}")
    print(f"Mean reprojection error: {result.mean_reproj_error:.6f} px")
    print(f"Recovered camera fx={cam.fx:.3f}, fy={cam.fy:.3f}, cx={cam.cx:.3f}, cy={cam.cy:.3f}")
    if result.gripper_se3_camera is not None:
        t = result.gripper_se3_camera.translation_xyz
        print(f"Recovered gripper->camera translation: [{t[0]:.4f}, {t[1]:.4f}, {t[2]:.4f}] m")


if __name__ == "__main__":
    main()



from __future__ import annotations

import unittest

import vision_calibration as vc


class HighLevelApiContractTest(unittest.TestCase):
    def test_dict_inputs_are_rejected_for_all_high_level_runners(self) -> None:
        cases = [
            (vc.run_planar_intrinsics, {"views": []}, vc.PlanarCalibrationConfig()),
            (
                vc.run_scheimpflug_intrinsics,
                {"views": []},
                vc.ScheimpflugIntrinsicsCalibrationConfig(),
            ),
            (
                vc.run_single_cam_handeye,
                {"views": []},
                vc.SingleCamHandeyeCalibrationConfig(),
            ),
            (
                vc.run_rig_extrinsics,
                {"num_cameras": 2, "views": []},
                vc.RigExtrinsicsCalibrationConfig(),
            ),
            (
                vc.run_rig_handeye,
                {"num_cameras": 2, "views": []},
                vc.RigHandeyeCalibrationConfig(),
            ),
            (
                vc.run_laserline_device,
                [],
                vc.LaserlineDeviceCalibrationConfig(),
            ),
        ]

        for fn, raw_input, cfg in cases:
            with self.subTest(runner=fn.__name__, mode="input"):
                with self.assertRaises(TypeError):
                    fn(raw_input, cfg)

    def test_dict_configs_are_rejected_for_all_high_level_runners(self) -> None:
        cases = [
            (vc.run_planar_intrinsics, vc.PlanarDataset(views=[])),
            (vc.run_scheimpflug_intrinsics, vc.PlanarDataset(views=[])),
            (vc.run_single_cam_handeye, vc.SingleCamHandeyeDataset(views=[])),
            (vc.run_rig_extrinsics, vc.RigExtrinsicsDataset(num_cameras=2, views=[])),
            (vc.run_rig_handeye, vc.RigHandeyeDataset(num_cameras=2, views=[])),
            (vc.run_laserline_device, vc.LaserlineDataset(views=[])),
        ]

        for fn, dataset in cases:
            with self.subTest(runner=fn.__name__, mode="config"):
                with self.assertRaises(TypeError):
                    fn(dataset, {"max_iters": 10})

    def test_typed_inputs_take_the_runtime_path(self) -> None:
        # Datasets are intentionally invalid (too few views), so the expected
        # failure mode is RuntimeError from Rust validation, not TypeError.
        cases = [
            (vc.run_planar_intrinsics, vc.PlanarDataset(views=[]), vc.PlanarCalibrationConfig()),
            (
                vc.run_scheimpflug_intrinsics,
                vc.PlanarDataset(views=[]),
                vc.ScheimpflugIntrinsicsCalibrationConfig(),
            ),
            (
                vc.run_single_cam_handeye,
                vc.SingleCamHandeyeDataset(views=[]),
                vc.SingleCamHandeyeCalibrationConfig(),
            ),
            (
                vc.run_rig_extrinsics,
                vc.RigExtrinsicsDataset(num_cameras=2, views=[]),
                vc.RigExtrinsicsCalibrationConfig(),
            ),
            (
                vc.run_rig_handeye,
                vc.RigHandeyeDataset(num_cameras=2, views=[]),
                vc.RigHandeyeCalibrationConfig(),
            ),
            (
                vc.run_laserline_device,
                vc.LaserlineDataset(views=[]),
                vc.LaserlineDeviceCalibrationConfig(),
            ),
        ]

        for fn, dataset, cfg in cases:
            with self.subTest(runner=fn.__name__):
                with self.assertRaises(RuntimeError):
                    fn(dataset, cfg)


if __name__ == "__main__":
    unittest.main()



"""Planar intrinsics calibration with deterministic synthetic data.

Python counterpart of:
`crates/vision-calibration/examples/planar_synthetic.rs`
"""

from __future__ import annotations

import vision_calibration as vc

from _common import (
    grid_points,
    make_pose,
    project_point_pinhole,
    transform_point,
)


def make_input() -> vc.PlanarDataset:
    intrinsics_gt = {"fx": 800.0, "fy": 780.0, "cx": 640.0, "cy": 360.0, "skew": 0.0}
    distortion_gt = {"k1": 0.05, "k2": -0.02, "k3": 0.0, "p1": 0.001, "p2": -0.001, "iters": 8}
    board = grid_points(cols=8, rows=6, spacing_m=0.04)

    poses = [
        make_pose(0.02 * i, -0.18 + 0.04 * i, -0.20 + 0.06 * i, -0.04 + 0.02 * i, 0.02 - 0.01 * i, 0.55 + 0.04 * i)
        for i in range(8)
    ]

    views: list[vc.PlanarView] = []
    for pose in poses:
        points_3d = []
        points_2d = []
        for p in board:
            p_cam = transform_point(pose, p)
            uv = project_point_pinhole(p_cam, intrinsics_gt, distortion_gt)
            if uv is None:
                continue
            points_3d.append(p)
            points_2d.append(uv)
        if len(points_2d) >= 4:
            views.append(vc.PlanarView(observation=vc.Observation(points_3d=points_3d, points_2d=points_2d)))

    return vc.PlanarDataset(views=views)


def main() -> None:
    print("=== Planar Intrinsics Calibration (Synthetic Data, Python) ===")
    payload = make_input()
    result = vc.run_planar_intrinsics(payload)
    k = result.camera.intrinsics
    d = result.camera.distortion

    print(f"Views: {len(payload.views)}")
    print(f"Final cost: {result.final_cost:.3e}")
    print(f"Mean reprojection error: {result.mean_reproj_error:.6f} px")
    print("Recovered intrinsics:")
    print(f"  fx={k.fx:.3f}, fy={k.fy:.3f}, cx={k.cx:.3f}, cy={k.cy:.3f}")
    print("Recovered distortion:")
    print(f"  k1={d.k1:.6f}, k2={d.k2:.6f}, p1={d.p1:.6f}, p2={d.p2:.6f}")


if __name__ == "__main__":
    main()



"""Public high-level models for :mod:`vision_calibration`.

These dataclasses provide a Python-native API surface with explicit fields,
defaults, and docstrings. Wrapper functions in :mod:`vision_calibration`
accept these models and convert them to the serde payloads expected by Rust.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Mapping, cast

from .types import HandEyeMode, LaserlineResidualType, RobustLoss

Vec2 = tuple[float, float]
Vec3 = tuple[float, float, float]
QuatXyzw = tuple[float, float, float, float]

_DEFAULT_INTRINSICS_FIX_MASK: dict[str, bool] = {
    "fx": False,
    "fy": False,
    "cx": False,
    "cy": False,
}
_DEFAULT_DISTORTION_FIX_MASK: dict[str, bool] = {
    "k1": False,
    "k2": False,
    "k3": True,
    "p1": False,
    "p2": False,
}
_DEFAULT_SENSOR_INIT: dict[str, float] = {
    "tilt_x": 0.0,
    "tilt_y": 0.0,
}
_DEFAULT_SCHEIMPFLUG_FIX_MASK: dict[str, bool] = {
    "tilt_x": False,
    "tilt_y": False,
}


def _as_floats(values: tuple[Any, ...] | list[Any], expected: int, name: str) -> tuple[float, ...]:
    if len(values) != expected:
        raise ValueError(f"{name} must have length {expected}, got {len(values)}")
    return tuple(float(v) for v in values)


def _as_vec2(values: tuple[Any, Any] | list[Any]) -> Vec2:
    x, y = _as_floats(values, 2, "Vec2")
    return (x, y)


def _as_vec3(values: tuple[Any, Any, Any] | list[Any]) -> Vec3:
    x, y, z = _as_floats(values, 3, "Vec3")
    return (x, y, z)


def _as_quat(values: tuple[Any, Any, Any, Any] | list[Any]) -> QuatXyzw:
    x, y, z, w = _as_floats(values, 4, "QuatXyzw")
    return (x, y, z, w)


def _payload_from_maybe_model(value: Any) -> Any:
    if hasattr(value, "to_payload"):
        return value.to_payload()
    return value


@dataclass(slots=True)
class Pose:
    """Rigid pose with quaternion rotation and translation.

    Parameters
    ----------
    rotation_xyzw:
        Quaternion as `(x, y, z, w)`.
    translation_xyz:
        Translation as `(x, y, z)` in meters.
    """

    rotation_xyzw: QuatXyzw = (0.0, 0.0, 0.0, 1.0)
    translation_xyz: Vec3 = (0.0, 0.0, 0.0)

    def __post_init__(self) -> None:
        self.rotation_xyzw = _as_quat(list(self.rotation_xyzw))
        self.translation_xyz = _as_vec3(list(self.translation_xyz))

    def to_payload(self) -> dict[str, list[float]]:
        """Convert to Rust/serde shape."""
        return {
            "rotation": [*self.rotation_xyzw],
            "translation": [*self.translation_xyz],
        }

    @classmethod
    def from_payload(cls, payload: Mapping[str, Any]) -> "Pose":
        """Build pose from Rust/serde shape."""
        return cls(
            rotation_xyzw=cast(tuple[float, float, float, float], tuple(payload["rotation"])),
            translation_xyz=cast(tuple[float, float, float], tuple(payload["translation"])),
        )


@dataclass(slots=True)
class Observation:
    """2D-3D correspondences for one camera view.

    Parameters
    ----------
    points_3d:
        Target points as `(x, y, z)` tuples.
    points_2d:
        Pixel points as `(u, v)` tuples.
    weights:
        Optional per-point non-negative weights.
    """

    points_3d: list[Vec3]
    points_2d: list[Vec2]
    weights: list[float] | None = None

    def __post_init__(self) -> None:
        self.points_3d = [_as_vec3(list(p)) for p in self.points_3d]
        self.points_2d = [_as_vec2(list(p)) for p in self.points_2d]
        if len(self.points_3d) != len(self.points_2d):
            raise ValueError(
                "points_3d and points_2d must have identical length "
                f"(got {len(self.points_3d)} vs {len(self.points_2d)})"
            )
        if self.weights is not None:
            if len(self.weights) != len(self.points_3d):
                raise ValueError(
                    "weights must match point count "
                    f"(got {len(self.weights)} vs {len(self.points_3d)})"
                )
            self.weights = [float(w) for w in self.weights]

    def to_payload(self) -> dict[str, Any]:
        """Convert to Rust/serde shape."""
        payload: dict[str, Any] = {
            "points_3d": [[x, y, z] for x, y, z in self.points_3d],
            "points_2d": [[u, v] for u, v in self.points_2d],
        }
        if self.weights is not None:
            payload["weights"] = list(self.weights)
        return payload

    @classmethod
    def from_payload(cls, payload: Mapping[str, Any]) -> "Observation":
        """Build observation from Rust/serde shape."""
        return cls(
            points_3d=[cast(Vec3, tuple(p)) for p in payload["points_3d"]],
            points_2d=[cast(Vec2, tuple(p)) for p in payload["points_2d"]],
            weights=cast(list[float] | None, payload.get("weights")),
        )


@dataclass(slots=True)
class PlanarView:
    """Planar calibration view."""

    observation: Observation

    def to_payload(self) -> dict[str, Any]:
        return {"obs": self.observation.to_payload(), "meta": None}


@dataclass(slots=True)
class PlanarDataset:
    """Planar calibration dataset."""

    views: list[PlanarView]

    def to_payload(self) -> dict[str, Any]:
        return {"views": [view.to_payload() for view in self.views]}


@dataclass(slots=True)
class SingleCamHandeyeView:
    """Single-camera hand-eye view."""

    observation: Observation
    base_se3_gripper: Pose

    def to_payload(self) -> dict[str, Any]:
        return {
            "obs": self.observation.to_payload(),
            "meta": {"base_se3_gripper": self.base_se3_gripper.to_payload()},
        }


@dataclass(slots=True)
class SingleCamHandeyeDataset:
    """Single-camera hand-eye dataset."""

    views: list[SingleCamHandeyeView]

    def to_payload(self) -> dict[str, Any]:
        return {"views": [view.to_payload() for view in self.views]}


@dataclass(slots=True)
class RigExtrinsicsView:
    """One frame in a multi-camera rig dataset.

    Parameters
    ----------
    cameras:
        Per-camera observation for this frame. Use `None` for a missing camera.
    """

    cameras: list[Observation | None]

    def to_payload(self) -> dict[str, Any]:
        return {
            "obs": {
                "cameras": [
                    None if obs is None else obs.to_payload()
                    for obs in self.cameras
                ]
            },
            "meta": None,
        }


@dataclass(slots=True)
class RigExtrinsicsDataset:
    """Multi-camera rig dataset for extrinsics calibration."""

    num_cameras: int
    views: list[RigExtrinsicsView]

    def to_payload(self) -> dict[str, Any]:
        return {
            "num_cameras": int(self.num_cameras),
            "views": [view.to_payload() for view in self.views],
        }


@dataclass(slots=True)
class RigHandeyeView:
    """One frame in rig hand-eye calibration."""

    cameras: list[Observation | None]
    base_se3_gripper: Pose

    def to_payload(self) -> dict[str, Any]:
        return {
            "obs": {
                "cameras": [
                    None if obs is None else obs.to_payload()
                    for obs in self.cameras
                ]
            },
            "meta": {"base_se3_gripper": self.base_se3_gripper.to_payload()},
        }


@dataclass(slots=True)
class RigHandeyeDataset:
    """Multi-camera rig hand-eye dataset."""

    num_cameras: int
    views: list[RigHandeyeView]

    def to_payload(self) -> dict[str, Any]:
        return {
            "num_cameras": int(self.num_cameras),
            "views": [view.to_payload() for view in self.views],
        }


@dataclass(slots=True)
class LaserlineView:
    """Single view for laserline-device calibration."""

    observation: Observation
    laser_pixels: list[Vec2]
    laser_weights: list[float] | None = None

    def __post_init__(self) -> None:
        self.laser_pixels = [_as_vec2(list(p)) for p in self.laser_pixels]
        if self.laser_weights is not None:
            if len(self.laser_weights) != len(self.laser_pixels):
                raise ValueError(
                    "laser_weights must match laser_pixels length "
                    f"(got {len(self.laser_weights)} vs {len(self.laser_pixels)})"
                )
            self.laser_weights = [float(w) for w in self.laser_weights]

    def to_payload(self) -> dict[str, Any]:
        return {
            "obs": self.observation.to_payload(),
            "meta": {
                "laser_pixels": [[u, v] for u, v in self.laser_pixels],
                "laser_weights": self.laser_weights,
            },
        }


@dataclass(slots=True)
class LaserlineDataset:
    """Dataset for laserline-device calibration."""

    views: list[LaserlineView]

    def to_payload(self) -> list[dict[str, Any]]:
        return [view.to_payload() for view in self.views]


@dataclass(slots=True)
class PlanarCalibrationConfig:
    """Configuration for planar intrinsics calibration."""

    init_iterations: int = 2
    fix_k3_in_init: bool = True
    fix_tangential_in_init: bool = False
    zero_skew: bool = True
    max_iters: int = 50
    verbosity: int = 0
    robust_loss: RobustLoss = "None"
    fix_intrinsics: dict[str, bool] = field(default_factory=lambda: dict(_DEFAULT_INTRINSICS_FIX_MASK))
    fix_distortion: dict[str, bool] = field(default_factory=lambda: dict(_DEFAULT_DISTORTION_FIX_MASK))
    fix_poses: list[int] = field(default_factory=list)

    def to_payload(self) -> dict[str, Any]:
        fix_intrinsics = dict(_DEFAULT_INTRINSICS_FIX_MASK)
        fix_intrinsics.update(self.fix_intrinsics)
        fix_distortion = dict(_DEFAULT_DISTORTION_FIX_MASK)
        fix_distortion.update(self.fix_distortion)
        return {
            "init_iterations": int(self.init_iterations),
            "fix_k3_in_init": bool(self.fix_k3_in_init),
            "fix_tangential_in_init": bool(self.fix_tangential_in_init),
            "zero_skew": bool(self.zero_skew),
            "max_iters": int(self.max_iters),
            "verbosity": int(self.verbosity),
            "robust_loss": cast(Any, self.robust_loss),
            "fix_intrinsics": fix_intrinsics,
            "fix_distortion": fix_distortion,
            "fix_poses": [int(i) for i in self.fix_poses],
        }

    @classmethod
    def from_mapping(cls, mapping: Mapping[str, Any]) -> "PlanarCalibrationConfig":
        cfg = cls()
        for key, value in mapping.items():
            if not hasattr(cfg, key):
                raise ValueError(f"unknown PlanarCalibrationConfig field: {key}")
            setattr(cfg, key, value)
        return cfg


@dataclass(slots=True)
class SingleCamHandeyeCalibrationConfig:
    """Configuration for single-camera hand-eye calibration."""

    intrinsics_init_iterations: int = 2
    fix_k3: bool = True
    fix_tangential: bool = False
    zero_skew: bool = True
    handeye_mode: HandEyeMode = "EyeInHand"
    min_motion_angle_deg: float = 5.0
    max_iters: int = 50
    verbosity: int = 0
    robust_loss: RobustLoss = "None"
    refine_robot_poses: bool = True
    robot_rot_sigma: float = 0.5 * 3.141592653589793 / 180.0
    robot_trans_sigma: float = 0.001

    def to_payload(self) -> dict[str, Any]:
        return {
            "intrinsics_init_iterations": int(self.intrinsics_init_iterations),
            "fix_k3": bool(self.fix_k3),
            "fix_tangential": bool(self.fix_tangential),
            "zero_skew": bool(self.zero_skew),
            "handeye_mode": self.handeye_mode,
            "min_motion_angle_deg": float(self.min_motion_angle_deg),
            "max_iters": int(self.max_iters),
            "verbosity": int(self.verbosity),
            "robust_loss": cast(Any, self.robust_loss),
            "refine_robot_poses": bool(self.refine_robot_poses),
            "robot_rot_sigma": float(self.robot_rot_sigma),
            "robot_trans_sigma": float(self.robot_trans_sigma),
        }

    @classmethod
    def from_mapping(cls, mapping: Mapping[str, Any]) -> "SingleCamHandeyeCalibrationConfig":
        cfg = cls()
        for key, value in mapping.items():
            if not hasattr(cfg, key):
                raise ValueError(f"unknown SingleCamHandeyeCalibrationConfig field: {key}")
            setattr(cfg, key, value)
        return cfg


@dataclass(slots=True)
class RigExtrinsicsCalibrationConfig:
    """Configuration for rig extrinsics calibration."""

    intrinsics_init_iterations: int = 2
    fix_k3: bool = True
    fix_tangential: bool = False
    zero_skew: bool = True
    reference_camera_idx: int = 0
    max_iters: int = 50
    verbosity: int = 0
    robust_loss: RobustLoss = "None"
    refine_intrinsics_in_rig_ba: bool = False
    fix_first_rig_pose: bool = True

    def to_payload(self) -> dict[str, Any]:
        return {
            "intrinsics_init_iterations": int(self.intrinsics_init_iterations),
            "fix_k3": bool(self.fix_k3),
            "fix_tangential": bool(self.fix_tangential),
            "zero_skew": bool(self.zero_skew),
            "reference_camera_idx": int(self.reference_camera_idx),
            "max_iters": int(self.max_iters),
            "verbosity": int(self.verbosity),
            "robust_loss": cast(Any, self.robust_loss),
            "refine_intrinsics_in_rig_ba": bool(self.refine_intrinsics_in_rig_ba),
            "fix_first_rig_pose": bool(self.fix_first_rig_pose),
        }

    @classmethod
    def from_mapping(cls, mapping: Mapping[str, Any]) -> "RigExtrinsicsCalibrationConfig":
        cfg = cls()
        for key, value in mapping.items():
            if not hasattr(cfg, key):
                raise ValueError(f"unknown RigExtrinsicsCalibrationConfig field: {key}")
            setattr(cfg, key, value)
        return cfg


@dataclass(slots=True)
class RigHandeyeIntrinsicsConfig:
    """Per-camera intrinsics initialization options."""

    init_iterations: int = 2
    fix_k3: bool = True
    fix_tangential: bool = False
    zero_skew: bool = True

    def to_payload(self) -> dict[str, Any]:
        return {
            "init_iterations": int(self.init_iterations),
            "fix_k3": bool(self.fix_k3),
            "fix_tangential": bool(self.fix_tangential),
            "zero_skew": bool(self.zero_skew),
        }

    @classmethod
    def from_mapping(cls, mapping: Mapping[str, Any]) -> "RigHandeyeIntrinsicsConfig":
        cfg = cls()
        for key, value in mapping.items():
            if not hasattr(cfg, key):
                raise ValueError(f"unknown RigHandeyeIntrinsicsConfig field: {key}")
            setattr(cfg, key, value)
        return cfg


@dataclass(slots=True)
class RigHandeyeRigConfig:
    """Rig frame / gauge options."""

    reference_camera_idx: int = 0
    refine_intrinsics_in_rig_ba: bool = False
    fix_first_rig_pose: bool = True

    def to_payload(self) -> dict[str, Any]:
        return {
            "reference_camera_idx": int(self.reference_camera_idx),
            "refine_intrinsics_in_rig_ba": bool(self.refine_intrinsics_in_rig_ba),
            "fix_first_rig_pose": bool(self.fix_first_rig_pose),
        }

    @classmethod
    def from_mapping(cls, mapping: Mapping[str, Any]) -> "RigHandeyeRigConfig":
        cfg = cls()
        for key, value in mapping.items():
            if not hasattr(cfg, key):
                raise ValueError(f"unknown RigHandeyeRigConfig field: {key}")
            setattr(cfg, key, value)
        return cfg


@dataclass(slots=True)
class RigHandeyeInitConfig:
    """Linear hand-eye initialization settings."""

    handeye_mode: HandEyeMode = "EyeInHand"
    min_motion_angle_deg: float = 5.0

    def to_payload(self) -> dict[str, Any]:
        return {
            "handeye_mode": self.handeye_mode,
            "min_motion_angle_deg": float(self.min_motion_angle_deg),
        }

    @classmethod
    def from_mapping(cls, mapping: Mapping[str, Any]) -> "RigHandeyeInitConfig":
        cfg = cls()
        for key, value in mapping.items():
            if not hasattr(cfg, key):
                raise ValueError(f"unknown RigHandeyeInitConfig field: {key}")
            setattr(cfg, key, value)
        return cfg


@dataclass(slots=True)
class RigHandeyeSolverConfig:
    """Shared nonlinear solver settings."""

    max_iters: int = 50
    verbosity: int = 0
    robust_loss: RobustLoss = "None"

    def to_payload(self) -> dict[str, Any]:
        return {
            "max_iters": int(self.max_iters),
            "verbosity": int(self.verbosity),
            "robust_loss": cast(Any, self.robust_loss),
        }

    @classmethod
    def from_mapping(cls, mapping: Mapping[str, Any]) -> "RigHandeyeSolverConfig":
        cfg = cls()
        for key, value in mapping.items():
            if not hasattr(cfg, key):
                raise ValueError(f"unknown RigHandeyeSolverConfig field: {key}")
            setattr(cfg, key, value)
        return cfg


@dataclass(slots=True)
class RigHandeyeBaConfig:
    """Final hand-eye bundle-adjustment options."""

    refine_robot_poses: bool = True
    robot_rot_sigma: float = 0.5 * 3.141592653589793 / 180.0
    robot_trans_sigma: float = 0.001
    refine_cam_se3_rig_in_handeye_ba: bool = False

    def to_payload(self) -> dict[str, Any]:
        return {
            "refine_robot_poses": bool(self.refine_robot_poses),
            "robot_rot_sigma": float(self.robot_rot_sigma),
            "robot_trans_sigma": float(self.robot_trans_sigma),
            "refine_cam_se3_rig_in_handeye_ba": bool(self.refine_cam_se3_rig_in_handeye_ba),
        }

    @classmethod
    def from_mapping(cls, mapping: Mapping[str, Any]) -> "RigHandeyeBaConfig":
        cfg = cls()
        for key, value in mapping.items():
            if not hasattr(cfg, key):
                raise ValueError(f"unknown RigHandeyeBaConfig field: {key}")
            setattr(cfg, key, value)
        return cfg


@dataclass(slots=True)
class RigHandeyeCalibrationConfig:
    """Configuration for rig hand-eye calibration."""

    intrinsics: RigHandeyeIntrinsicsConfig = field(default_factory=RigHandeyeIntrinsicsConfig)
    rig: RigHandeyeRigConfig = field(default_factory=RigHandeyeRigConfig)
    handeye_init: RigHandeyeInitConfig = field(default_factory=RigHandeyeInitConfig)
    solver: RigHandeyeSolverConfig = field(default_factory=RigHandeyeSolverConfig)
    handeye_ba: RigHandeyeBaConfig = field(default_factory=RigHandeyeBaConfig)

    def to_payload(self) -> dict[str, Any]:
        return {
            "intrinsics": self.intrinsics.to_payload(),
            "rig": self.rig.to_payload(),
            "handeye_init": self.handeye_init.to_payload(),
            "solver": self.solver.to_payload(),
            "handeye_ba": self.handeye_ba.to_payload(),
        }

    @classmethod
    def from_mapping(cls, mapping: Mapping[str, Any]) -> "RigHandeyeCalibrationConfig":
        cfg = cls()
        for key, value in mapping.items():
            if key == "intrinsics":
                cfg.intrinsics = (
                    value
                    if isinstance(value, RigHandeyeIntrinsicsConfig)
                    else RigHandeyeIntrinsicsConfig.from_mapping(cast(Mapping[str, Any], value))
                )
            elif key == "rig":
                cfg.rig = (
                    value
                    if isinstance(value, RigHandeyeRigConfig)
                    else RigHandeyeRigConfig.from_mapping(cast(Mapping[str, Any], value))
                )
            elif key == "handeye_init":
                cfg.handeye_init = (
                    value
                    if isinstance(value, RigHandeyeInitConfig)
                    else RigHandeyeInitConfig.from_mapping(cast(Mapping[str, Any], value))
                )
            elif key == "solver":
                cfg.solver = (
                    value
                    if isinstance(value, RigHandeyeSolverConfig)
                    else RigHandeyeSolverConfig.from_mapping(cast(Mapping[str, Any], value))
                )
            elif key == "handeye_ba":
                cfg.handeye_ba = (
                    value
                    if isinstance(value, RigHandeyeBaConfig)
                    else RigHandeyeBaConfig.from_mapping(cast(Mapping[str, Any], value))
                )
            else:
                raise ValueError(f"unknown RigHandeyeCalibrationConfig field: {key}")
        return cfg


@dataclass(slots=True)
class LaserlineDeviceInitConfig:
    """Initialization options for laserline-device calibration."""

    iterations: int = 2
    fix_k3: bool = True
    fix_tangential: bool = False
    zero_skew: bool = True
    sensor_init: dict[str, float] = field(default_factory=lambda: dict(_DEFAULT_SENSOR_INIT))

    def to_payload(self) -> dict[str, Any]:
        sensor_init = dict(_DEFAULT_SENSOR_INIT)
        sensor_init.update(self.sensor_init)
        return {
            "iterations": int(self.iterations),
            "fix_k3": bool(self.fix_k3),
            "fix_tangential": bool(self.fix_tangential),
            "zero_skew": bool(self.zero_skew),
            "sensor_init": sensor_init,
        }

    @classmethod
    def from_mapping(cls, mapping: Mapping[str, Any]) -> "LaserlineDeviceInitConfig":
        cfg = cls()
        for key, value in mapping.items():
            if not hasattr(cfg, key):
                raise ValueError(f"unknown LaserlineDeviceInitConfig field: {key}")
            setattr(cfg, key, value)
        return cfg


@dataclass(slots=True)
class LaserlineDeviceSolverConfig:
    """Nonlinear solver options for laserline-device calibration."""

    max_iters: int = 50
    verbosity: int = 0

    def to_payload(self) -> dict[str, Any]:
        return {
            "max_iters": int(self.max_iters),
            "verbosity": int(self.verbosity),
        }

    @classmethod
    def from_mapping(cls, mapping: Mapping[str, Any]) -> "LaserlineDeviceSolverConfig":
        cfg = cls()
        for key, value in mapping.items():
            if not hasattr(cfg, key):
                raise ValueError(f"unknown LaserlineDeviceSolverConfig field: {key}")
            setattr(cfg, key, value)
        return cfg


@dataclass(slots=True)
class LaserlineDeviceOptimizeConfig:
    """Bundle-adjustment options for laserline-device calibration."""

    calib_loss: RobustLoss = field(default_factory=lambda: {"Huber": {"scale": 1.0}})
    laser_loss: RobustLoss = field(default_factory=lambda: {"Huber": {"scale": 0.01}})
    calib_weight: float = 1.0
    laser_weight: float = 1.0
    fix_intrinsics: bool = False
    fix_distortion: bool = False
    fix_k3: bool = True
    fix_sensor: bool = True
    fix_poses: list[int] = field(default_factory=lambda: [0])
    fix_plane: bool = False
    laser_residual_type: LaserlineResidualType = "LineDistNormalized"

    def to_payload(self) -> dict[str, Any]:
        return {
            "calib_loss": cast(Any, self.calib_loss),
            "laser_loss": cast(Any, self.laser_loss),
            "calib_weight": float(self.calib_weight),
            "laser_weight": float(self.laser_weight),
            "fix_intrinsics": bool(self.fix_intrinsics),
            "fix_distortion": bool(self.fix_distortion),
            "fix_k3": bool(self.fix_k3),
            "fix_sensor": bool(self.fix_sensor),
            "fix_poses": [int(i) for i in self.fix_poses],
            "fix_plane": bool(self.fix_plane),
            "laser_residual_type": self.laser_residual_type,
        }

    @classmethod
    def from_mapping(cls, mapping: Mapping[str, Any]) -> "LaserlineDeviceOptimizeConfig":
        cfg = cls()
        for key, value in mapping.items():
            if not hasattr(cfg, key):
                raise ValueError(f"unknown LaserlineDeviceOptimizeConfig field: {key}")
            setattr(cfg, key, value)
        return cfg


@dataclass(slots=True)
class LaserlineDeviceCalibrationConfig:
    """Configuration for laserline-device calibration."""

    init: LaserlineDeviceInitConfig = field(default_factory=LaserlineDeviceInitConfig)
    solver: LaserlineDeviceSolverConfig = field(default_factory=LaserlineDeviceSolverConfig)
    optimize: LaserlineDeviceOptimizeConfig = field(default_factory=LaserlineDeviceOptimizeConfig)

    def to_payload(self) -> dict[str, Any]:
        return {
            "init": self.init.to_payload(),
            "solver": self.solver.to_payload(),
            "optimize": self.optimize.to_payload(),
        }

    @classmethod
    def from_mapping(cls, mapping: Mapping[str, Any]) -> "LaserlineDeviceCalibrationConfig":
        cfg = cls()
        for key, value in mapping.items():
            if key == "init":
                cfg.init = (
                    value
                    if isinstance(value, LaserlineDeviceInitConfig)
                    else LaserlineDeviceInitConfig.from_mapping(cast(Mapping[str, Any], value))
                )
            elif key == "solver":
                cfg.solver = (
                    value
                    if isinstance(value, LaserlineDeviceSolverConfig)
                    else LaserlineDeviceSolverConfig.from_mapping(cast(Mapping[str, Any], value))
                )
            elif key == "optimize":
                cfg.optimize = (
                    value
                    if isinstance(value, LaserlineDeviceOptimizeConfig)
                    else LaserlineDeviceOptimizeConfig.from_mapping(cast(Mapping[str, Any], value))
                )
            else:
                raise ValueError(f"unknown LaserlineDeviceCalibrationConfig field: {key}")
        return cfg


@dataclass(slots=True)
class ScheimpflugIntrinsicsCalibrationConfig:
    """Configuration for planar Scheimpflug intrinsics calibration."""

    init_iterations: int = 2
    fix_k3_in_init: bool = True
    zero_skew: bool = True
    max_iters: int = 120
    verbosity: int = 0
    robust_loss: RobustLoss = field(default_factory=lambda: "None")
    fix_intrinsics: dict[str, bool] = field(default_factory=lambda: dict(_DEFAULT_INTRINSICS_FIX_MASK))
    fix_distortion: dict[str, bool] = field(
        default_factory=lambda: {
            "k1": False,
            "k2": False,
            "k3": True,
            "p1": True,
            "p2": True,
        }
    )
    fix_scheimpflug: dict[str, bool] = field(default_factory=lambda: dict(_DEFAULT_SCHEIMPFLUG_FIX_MASK))
    fix_first_pose: bool = True

    def to_payload(self) -> dict[str, Any]:
        fix_intrinsics = dict(_DEFAULT_INTRINSICS_FIX_MASK)
        fix_intrinsics.update(self.fix_intrinsics)
        fix_distortion = {
            "k1": False,
            "k2": False,
            "k3": True,
            "p1": True,
            "p2": True,
        }
        fix_distortion.update(self.fix_distortion)
        fix_scheimpflug = dict(_DEFAULT_SCHEIMPFLUG_FIX_MASK)
        fix_scheimpflug.update(self.fix_scheimpflug)
        return {
            "init_iterations": int(self.init_iterations),
            "fix_k3_in_init": bool(self.fix_k3_in_init),
            "zero_skew": bool(self.zero_skew),
            "max_iters": int(self.max_iters),
            "verbosity": int(self.verbosity),
            "robust_loss": cast(Any, self.robust_loss),
            "fix_intrinsics": fix_intrinsics,
            "fix_distortion": fix_distortion,
            "fix_scheimpflug": fix_scheimpflug,
            "fix_first_pose": bool(self.fix_first_pose),
        }

    @classmethod
    def from_mapping(cls, mapping: Mapping[str, Any]) -> "ScheimpflugIntrinsicsCalibrationConfig":
        cfg = cls()
        for key, value in mapping.items():
            if not hasattr(cfg, key):
                raise ValueError(f"unknown ScheimpflugIntrinsicsCalibrationConfig field: {key}")
            setattr(cfg, key, value)
        return cfg


@dataclass(slots=True)
class PinholeIntrinsics:
    """Typed pinhole intrinsics model."""

    fx: float
    fy: float
    cx: float
    cy: float
    skew: float

    def to_payload(self) -> dict[str, float]:
        """Convert to serde payload shape."""
        return {
            "fx": float(self.fx),
            "fy": float(self.fy),
            "cx": float(self.cx),
            "cy": float(self.cy),
            "skew": float(self.skew),
        }

    @classmethod
    def from_payload(cls, payload: Mapping[str, Any]) -> "PinholeIntrinsics":
        """Parse intrinsics from serde payload shape."""
        return cls(
            fx=float(payload["fx"]),
            fy=float(payload["fy"]),
            cx=float(payload["cx"]),
            cy=float(payload["cy"]),
            skew=float(payload["skew"]),
        )


@dataclass(slots=True)
class BrownConradyDistortion:
    """Typed Brown-Conrady distortion model."""

    k1: float
    k2: float
    k3: float
    p1: float
    p2: float
    iters: int

    def to_payload(self) -> dict[str, float | int]:
        """Convert to serde payload shape."""
        return {
            "k1": float(self.k1),
            "k2": float(self.k2),
            "k3": float(self.k3),
            "p1": float(self.p1),
            "p2": float(self.p2),
            "iters": int(self.iters),
        }

    @classmethod
    def from_payload(cls, payload: Mapping[str, Any]) -> "BrownConradyDistortion":
        """Parse distortion from serde payload shape."""
        return cls(
            k1=float(payload["k1"]),
            k2=float(payload["k2"]),
            k3=float(payload["k3"]),
            p1=float(payload["p1"]),
            p2=float(payload["p2"]),
            iters=int(payload["iters"]),
        )


@dataclass(slots=True)
class ScheimpflugSensor:
    """Typed Scheimpflug sensor tilt model."""

    tilt_x: float
    tilt_y: float

    def to_payload(self) -> dict[str, float]:
        """Convert to serde payload shape."""
        return {
            "tilt_x": float(self.tilt_x),
            "tilt_y": float(self.tilt_y),
        }

    @classmethod
    def from_payload(cls, payload: Mapping[str, Any]) -> "ScheimpflugSensor":
        """Parse Scheimpflug tilt, accepting legacy `tau_x`/`tau_y` aliases."""
        tilt_x = payload.get("tilt_x", payload.get("tau_x"))
        tilt_y = payload.get("tilt_y", payload.get("tau_y"))
        if tilt_x is None:
            raise ValueError("Scheimpflug sensor payload missing tilt_x/tau_x")
        if tilt_y is None:
            raise ValueError("Scheimpflug sensor payload missing tilt_y/tau_y")
        return cls(tilt_x=float(tilt_x), tilt_y=float(tilt_y))


@dataclass(slots=True)
class PinholeBrownConradyCamera:
    """Typed pinhole camera model with Brown-Conrady distortion."""

    intrinsics: PinholeIntrinsics
    distortion: BrownConradyDistortion

    def to_payload(self) -> dict[str, Any]:
        """Convert to the `PinholeCamera` serde payload shape."""
        return {
            "k": self.intrinsics.to_payload(),
            "dist": self.distortion.to_payload(),
        }

    @classmethod
    def from_payload(cls, payload: Mapping[str, Any]) -> "PinholeBrownConradyCamera":
        """Parse from either `PinholeCamera` or `CameraParams`-style payload."""
        if "k" in payload and "dist" in payload:
            return cls(
                intrinsics=PinholeIntrinsics.from_payload(cast(Mapping[str, Any], payload["k"])),
                distortion=BrownConradyDistortion.from_payload(
                    cast(Mapping[str, Any], payload["dist"])
                ),
            )

        if "intrinsics" in payload and "distortion" in payload:
            intrinsics_payload = cast(Mapping[str, Any], payload["intrinsics"])
            distortion_payload = cast(Mapping[str, Any], payload["distortion"])

            intrinsics_type = intrinsics_payload.get("type")
            if intrinsics_type not in (None, "fx_fy_cx_cy_skew"):
                raise ValueError(
                    f"unsupported intrinsics payload type for pinhole camera: {intrinsics_type!r}"
                )

            distortion_type = distortion_payload.get("type")
            if distortion_type not in (None, "brown_conrady5"):
                raise ValueError(
                    f"unsupported distortion payload type for pinhole camera: {distortion_type!r}"
                )

            return cls(
                intrinsics=PinholeIntrinsics.from_payload(intrinsics_payload),
                distortion=BrownConradyDistortion.from_payload(distortion_payload),
            )

        raise ValueError("camera payload missing expected intrinsics/distortion fields")


@dataclass(slots=True)
class PinholeBrownConradyScheimpflugCamera:
    """Typed pinhole + Brown-Conrady + Scheimpflug camera model."""

    intrinsics: PinholeIntrinsics
    distortion: BrownConradyDistortion
    sensor: ScheimpflugSensor

    def to_payload(self) -> dict[str, Any]:
        """Convert to `CameraParams` serde payload shape."""
        return {
            "projection": {"type": "pinhole"},
            "distortion": {"type": "brown_conrady5", **self.distortion.to_payload()},
            "sensor": {"type": "scheimpflug", **self.sensor.to_payload()},
            "intrinsics": {"type": "fx_fy_cx_cy_skew", **self.intrinsics.to_payload()},
        }

    @classmethod
    def from_payload(cls, payload: Mapping[str, Any]) -> "PinholeBrownConradyScheimpflugCamera":
        """Parse from `CameraParams` payload with Scheimpflug sensor."""
        intrinsics_payload = cast(Mapping[str, Any], payload["intrinsics"])
        distortion_payload = cast(Mapping[str, Any], payload["distortion"])
        sensor_payload = cast(Mapping[str, Any], payload["sensor"])

        intrinsics_type = intrinsics_payload.get("type")
        if intrinsics_type not in (None, "fx_fy_cx_cy_skew"):
            raise ValueError(
                f"unsupported intrinsics payload type for Scheimpflug camera: {intrinsics_type!r}"
            )

        distortion_type = distortion_payload.get("type")
        if distortion_type not in (None, "brown_conrady5"):
            raise ValueError(
                f"unsupported distortion payload type for Scheimpflug camera: {distortion_type!r}"
            )

        sensor_type = sensor_payload.get("type")
        if sensor_type not in (None, "scheimpflug"):
            raise ValueError(
                f"unsupported sensor payload type for Scheimpflug camera: {sensor_type!r}"
            )

        return cls(
            intrinsics=PinholeIntrinsics.from_payload(intrinsics_payload),
            distortion=BrownConradyDistortion.from_payload(distortion_payload),
            sensor=ScheimpflugSensor.from_payload(sensor_payload),
        )


@dataclass(slots=True)
class LaserlinePlane:
    """Typed laser plane representation."""

    normal_xyz: Vec3
    distance: float

    def __post_init__(self) -> None:
        self.normal_xyz = _as_vec3(list(self.normal_xyz))
        self.distance = float(self.distance)

    @classmethod
    def from_payload(cls, payload: Mapping[str, Any]) -> "LaserlinePlane":
        """Parse laser plane from serde payload."""
        normal_value = payload["normal"]
        if isinstance(normal_value, Mapping):
            if "coords" in normal_value:
                normal_value = normal_value["coords"]
            elif "data" in normal_value:
                normal_value = normal_value["data"]
            else:
                raise ValueError("unsupported laser plane normal payload shape")
        return cls(
            normal_xyz=cast(Vec3, tuple(normal_value)),
            distance=float(payload["distance"]),
        )


@dataclass(slots=True)
class LaserlineEstimateParams:
    """Typed parameter payload for laserline optimization results."""

    intrinsics: PinholeIntrinsics
    distortion: BrownConradyDistortion
    sensor: ScheimpflugSensor
    poses: list[Pose]
    plane: LaserlinePlane

    @classmethod
    def from_payload(cls, payload: Mapping[str, Any]) -> "LaserlineEstimateParams":
        return cls(
            intrinsics=PinholeIntrinsics.from_payload(
                cast(Mapping[str, Any], payload["intrinsics"])
            ),
            distortion=BrownConradyDistortion.from_payload(
                cast(Mapping[str, Any], payload["distortion"])
            ),
            sensor=ScheimpflugSensor.from_payload(cast(Mapping[str, Any], payload["sensor"])),
            poses=[
                Pose.from_payload(cast(Mapping[str, Any], pose))
                for pose in cast(list[Any], payload["poses"])
            ],
            plane=LaserlinePlane.from_payload(cast(Mapping[str, Any], payload["plane"])),
        )


@dataclass(slots=True)
class LaserlineEstimate:
    """Typed laserline optimizer estimate payload."""

    params: LaserlineEstimateParams
    final_cost: float

    @classmethod
    def from_payload(cls, payload: Mapping[str, Any]) -> "LaserlineEstimate":
        report = cast(Mapping[str, Any], payload["report"])
        return cls(
            params=LaserlineEstimateParams.from_payload(
                cast(Mapping[str, Any], payload["params"])
            ),
            final_cost=float(report["final_cost"]),
        )


@dataclass(slots=True)
class LaserlineStats:
    """Typed summary statistics for laserline results."""

    mean_reproj_error: float
    mean_laser_error: float
    per_view_reproj_errors: list[float]
    per_view_laser_errors: list[float]

    @classmethod
    def from_payload(cls, payload: Mapping[str, Any]) -> "LaserlineStats":
        return cls(
            mean_reproj_error=float(payload["mean_reproj_error"]),
            mean_laser_error=float(payload["mean_laser_error"]),
            per_view_reproj_errors=[
                float(v) for v in cast(list[Any], payload["per_view_reproj_errors"])
            ],
            per_view_laser_errors=[
                float(v) for v in cast(list[Any], payload["per_view_laser_errors"])
            ],
        )


@dataclass(slots=True)
class PlanarCalibrationResult:
    """Result from :func:`vision_calibration.run_planar_intrinsics`."""

    camera: PinholeBrownConradyCamera
    camera_se3_target: list[Pose]
    final_cost: float
    mean_reproj_error: float
    per_cam_reproj_errors: list[float]

    @classmethod
    def from_payload(cls, payload: Mapping[str, Any]) -> "PlanarCalibrationResult":
        params = cast(Mapping[str, Any], payload["params"])
        report = cast(Mapping[str, Any], payload["report"])
        poses = [
            Pose.from_payload(cast(Mapping[str, Any], p))
            for p in cast(list[Any], params["camera_se3_target"])
        ]
        return cls(
            camera=PinholeBrownConradyCamera.from_payload(
                cast(Mapping[str, Any], params["camera"])
            ),
            camera_se3_target=poses,
            final_cost=float(report["final_cost"]),
            mean_reproj_error=float(payload["mean_reproj_error"]),
            per_cam_reproj_errors=[
                float(v) for v in cast(list[Any], payload["per_cam_reproj_errors"])
            ],
        )


@dataclass(slots=True)
class SingleCamHandeyeResult:
    """Result from :func:`vision_calibration.run_single_cam_handeye`."""

    camera: PinholeBrownConradyCamera
    handeye_mode: HandEyeMode
    gripper_se3_camera: Pose | None
    camera_se3_base: Pose | None
    base_se3_target: Pose | None
    gripper_se3_target: Pose | None
    robot_deltas: list[list[float]] | None
    mean_reproj_error: float
    per_cam_reproj_errors: list[float]

    @classmethod
    def from_payload(cls, payload: Mapping[str, Any]) -> "SingleCamHandeyeResult":
        def _pose(name: str) -> Pose | None:
            value = payload.get(name)
            if value is None:
                return None
            return Pose.from_payload(cast(Mapping[str, Any], value))

        return cls(
            camera=PinholeBrownConradyCamera.from_payload(
                cast(Mapping[str, Any], payload["camera"])
            ),
            handeye_mode=cast(HandEyeMode, payload["handeye_mode"]),
            gripper_se3_camera=_pose("gripper_se3_camera"),
            camera_se3_base=_pose("camera_se3_base"),
            base_se3_target=_pose("base_se3_target"),
            gripper_se3_target=_pose("gripper_se3_target"),
            robot_deltas=cast(list[list[float]] | None, payload.get("robot_deltas")),
            mean_reproj_error=float(payload["mean_reproj_error"]),
            per_cam_reproj_errors=[float(v) for v in cast(list[Any], payload["per_cam_reproj_errors"])],
        )


@dataclass(slots=True)
class RigExtrinsicsResult:
    """Result from :func:`vision_calibration.run_rig_extrinsics`."""

    cameras: list[PinholeBrownConradyCamera]
    cam_se3_rig: list[Pose]
    mean_reproj_error: float
    per_cam_reproj_errors: list[float]

    @classmethod
    def from_payload(cls, payload: Mapping[str, Any]) -> "RigExtrinsicsResult":
        return cls(
            cameras=[
                PinholeBrownConradyCamera.from_payload(cast(Mapping[str, Any], c))
                for c in cast(list[Any], payload["cameras"])
            ],
            cam_se3_rig=[Pose.from_payload(cast(Mapping[str, Any], p)) for p in cast(list[Any], payload["cam_se3_rig"])],
            mean_reproj_error=float(payload["mean_reproj_error"]),
            per_cam_reproj_errors=[float(v) for v in cast(list[Any], payload["per_cam_reproj_errors"])],
        )


@dataclass(slots=True)
class RigHandeyeResult:
    """Result from :func:`vision_calibration.run_rig_handeye`."""

    cameras: list[PinholeBrownConradyCamera]
    cam_se3_rig: list[Pose]
    handeye_mode: HandEyeMode
    gripper_se3_rig: Pose | None
    rig_se3_base: Pose | None
    base_se3_target: Pose | None
    gripper_se3_target: Pose | None
    robot_deltas: list[list[float]] | None
    mean_reproj_error: float
    per_cam_reproj_errors: list[float]

    @classmethod
    def from_payload(cls, payload: Mapping[str, Any]) -> "RigHandeyeResult":
        def _pose(name: str) -> Pose | None:
            value = payload.get(name)
            if value is None:
                return None
            return Pose.from_payload(cast(Mapping[str, Any], value))

        return cls(
            cameras=[
                PinholeBrownConradyCamera.from_payload(cast(Mapping[str, Any], c))
                for c in cast(list[Any], payload["cameras"])
            ],
            cam_se3_rig=[Pose.from_payload(cast(Mapping[str, Any], p)) for p in cast(list[Any], payload["cam_se3_rig"])],
            handeye_mode=cast(HandEyeMode, payload["handeye_mode"]),
            gripper_se3_rig=_pose("gripper_se3_rig"),
            rig_se3_base=_pose("rig_se3_base"),
            base_se3_target=_pose("base_se3_target"),
            gripper_se3_target=_pose("gripper_se3_target"),
            robot_deltas=cast(list[list[float]] | None, payload.get("robot_deltas")),
            mean_reproj_error=float(payload["mean_reproj_error"]),
            per_cam_reproj_errors=[float(v) for v in cast(list[Any], payload["per_cam_reproj_errors"])],
        )


@dataclass(slots=True)
class LaserlineDeviceResult:
    """Result from :func:`vision_calibration.run_laserline_device`."""

    estimate: LaserlineEstimate
    stats: LaserlineStats
    mean_reproj_error: float
    per_cam_reproj_errors: list[float]

    @property
    def mean_laser_error(self) -> float:
        """Mean laser residual (units depend on residual type)."""
        return self.stats.mean_laser_error

    @classmethod
    def from_payload(cls, payload: Mapping[str, Any]) -> "LaserlineDeviceResult":
        return cls(
            estimate=LaserlineEstimate.from_payload(
                cast(Mapping[str, Any], payload["estimate"])
            ),
            stats=LaserlineStats.from_payload(cast(Mapping[str, Any], payload["stats"])),
            mean_reproj_error=float(payload["mean_reproj_error"]),
            per_cam_reproj_errors=[float(v) for v in cast(list[Any], payload["per_cam_reproj_errors"])],
        )


@dataclass(slots=True)
class ScheimpflugIntrinsicsResult:
    """Result from :func:`vision_calibration.run_scheimpflug_intrinsics`."""

    camera: PinholeBrownConradyScheimpflugCamera
    camera_se3_target: list[Pose]
    final_cost: float
    mean_reproj_error: float
    per_cam_reproj_errors: list[float]

    @classmethod
    def from_payload(cls, payload: Mapping[str, Any]) -> "ScheimpflugIntrinsicsResult":
        params = cast(Mapping[str, Any], payload["params"])
        report = cast(Mapping[str, Any], payload["report"])
        return cls(
            camera=PinholeBrownConradyScheimpflugCamera.from_payload(
                cast(Mapping[str, Any], params["camera"])
            ),
            camera_se3_target=[
                Pose.from_payload(cast(Mapping[str, Any], pose))
                for pose in cast(list[Any], params["camera_se3_target"])
            ],
            final_cost=float(report["final_cost"]),
            mean_reproj_error=float(payload["mean_reproj_error"]),
            per_cam_reproj_errors=[float(v) for v in cast(list[Any], payload["per_cam_reproj_errors"])],
        )


def normalize_input_payload(input_value: Any) -> Any:
    """Convert high-level model inputs to serde payloads."""
    return _payload_from_maybe_model(input_value)



"""Stereo rig calibration from real ChArUco images.

Python counterpart of:
`crates/vision-calibration/examples/stereo_charuco_session.rs`
"""

from __future__ import annotations

from pathlib import Path

import calib_targets
import numpy as np
from PIL import Image
import vision_calibration as vc

from _common import baseline_m, repo_root

BOARD_ROWS = 22
BOARD_COLS = 22
BOARD_CELL_SIZE_M = 0.00135
BOARD_MARKER_SIZE_REL = 0.75
BOARD_DICTIONARY = "DICT_4X4_1000"


def load_gray(path: Path) -> np.ndarray:
    return np.asarray(Image.open(path).convert("L"), dtype=np.uint8)


def list_pair_suffixes(left_dir: Path, right_dir: Path) -> list[str]:
    left = {p.name[len("Cam1_") :] for p in left_dir.glob("Cam1_*.png")}
    right = {p.name[len("Cam2_") :] for p in right_dir.glob("Cam2_*.png")}
    return sorted(left.intersection(right))


def make_charuco_params() -> calib_targets.CharucoDetectorParams:
    board = calib_targets.CharucoBoardSpec(
        rows=BOARD_ROWS,
        cols=BOARD_COLS,
        cell_size=BOARD_CELL_SIZE_M,
        marker_size_rel=BOARD_MARKER_SIZE_REL,
        dictionary=BOARD_DICTIONARY,
        marker_layout=calib_targets.MarkerLayout.OPENCV_CHARUCO,
    )
    return calib_targets.CharucoDetectorParams(
        board=board,
        graph=calib_targets.GridGraphParams(max_spacing_pix=120.0),
    )


def detect_charuco_observation(
    path: Path,
    params: calib_targets.CharucoDetectorParams,
) -> vc.Observation | None:
    image = load_gray(path)
    try:
        detection = calib_targets.detect_charuco(
            image,
            chess_cfg=calib_targets.ChessConfig(),
            params=params,
        )
    except Exception:
        return None

    points_3d: list[tuple[float, float, float]] = []
    points_2d: list[tuple[float, float]] = []
    for corner in detection.detection.corners:
        target = corner.target_position
        if target is None:
            continue
        points_3d.append((float(target[0]), float(target[1]), 0.0))
        points_2d.append((float(corner.position[0]), float(corner.position[1])))
    if len(points_2d) < 4:
        return None
    return vc.Observation(points_3d=points_3d, points_2d=points_2d)


def load_dataset() -> vc.RigExtrinsicsDataset:
    data_dir = repo_root() / "data/stereo_charuco"
    left_dir = data_dir / "cam1"
    right_dir = data_dir / "cam2"
    if not left_dir.exists() or not right_dir.exists():
        raise FileNotFoundError(f"stereo_charuco dataset not found: {data_dir}")

    params = make_charuco_params()
    views: list[vc.RigExtrinsicsView] = []
    skipped = 0
    left_ok = 0
    right_ok = 0
    for suffix in list_pair_suffixes(left_dir, right_dir):
        left = detect_charuco_observation(left_dir / f"Cam1_{suffix}", params)
        right = detect_charuco_observation(right_dir / f"Cam2_{suffix}", params)
        if left is None and right is None:
            skipped += 1
            continue
        left_ok += int(left is not None)
        right_ok += int(right is not None)
        views.append(vc.RigExtrinsicsView(cameras=[left, right]))

    if len(views) < 3:
        raise RuntimeError(f"need at least 3 usable views, got {len(views)}")
    print(f"Loaded {len(views)} views ({skipped} skipped), usable: left={left_ok}, right={right_ok}")
    return vc.RigExtrinsicsDataset(num_cameras=2, views=views)


def main() -> None:
    print("=== Stereo Rig Calibration Session (ChArUco, Python) ===")
    dataset = load_dataset()
    result = vc.run_rig_extrinsics(dataset)

    print(f"Mean reprojection error: {result.mean_reproj_error:.6f} px")
    print(f"Per-camera reprojection: {result.per_cam_reproj_errors}")
    print(f"Estimated baseline: {baseline_m(result.cam_se3_rig) * 1000.0:.3f} mm")


if __name__ == "__main__":
    main()



from __future__ import annotations

import math
import unittest

import vision_calibration as vc


def _rotation_from_euler_xyz(ax: float, ay: float, az: float) -> list[list[float]]:
    sx, cx = math.sin(ax), math.cos(ax)
    sy, cy = math.sin(ay), math.cos(ay)
    sz, cz = math.sin(az), math.cos(az)
    return [
        [cz * cy, cz * sy * sx - sz * cx, cz * sy * cx + sz * sx],
        [sz * cy, sz * sy * sx + cz * cx, sz * sy * cx - cz * sx],
        [-sy, cy * sx, cy * cx],
    ]


def _quat_xyzw_from_rot(r: list[list[float]]) -> list[float]:
    trace = r[0][0] + r[1][1] + r[2][2]
    if trace > 0.0:
        s = math.sqrt(trace + 1.0) * 2.0
        return [
            (r[2][1] - r[1][2]) / s,
            (r[0][2] - r[2][0]) / s,
            (r[1][0] - r[0][1]) / s,
            0.25 * s,
        ]
    raise ValueError("unexpected non-positive trace in test rotation")


def _make_pose(ax: float, ay: float, az: float, tx: float, ty: float, tz: float) -> dict:
    return {
        "rotation": _quat_xyzw_from_rot(_rotation_from_euler_xyz(ax, ay, az)),
        "translation": [tx, ty, tz],
    }


def _pose_to_rt(pose: dict) -> tuple[list[list[float]], list[float]]:
    x, y, z, w = pose["rotation"]
    tx, ty, tz = pose["translation"]
    xx = x * x
    yy = y * y
    zz = z * z
    xy = x * y
    xz = x * z
    yz = y * z
    wx = w * x
    wy = w * y
    wz = w * z
    return (
        [
            [1.0 - 2.0 * (yy + zz), 2.0 * (xy - wz), 2.0 * (xz + wy)],
            [2.0 * (xy + wz), 1.0 - 2.0 * (xx + zz), 2.0 * (yz - wx)],
            [2.0 * (xz - wy), 2.0 * (yz + wx), 1.0 - 2.0 * (xx + yy)],
        ],
        [tx, ty, tz],
    )


def _transform_point(pose: dict, point: list[float]) -> list[float]:
    r, t = _pose_to_rt(pose)
    return [
        r[0][0] * point[0] + r[0][1] * point[1] + r[0][2] * point[2] + t[0],
        r[1][0] * point[0] + r[1][1] * point[1] + r[1][2] * point[2] + t[1],
        r[2][0] * point[0] + r[2][1] * point[1] + r[2][2] * point[2] + t[2],
    ]


def _project_point_pinhole(point_cam: list[float]) -> list[float] | None:
    x, y, z = point_cam
    if z <= 0.0:
        return None
    return [640.0 + 800.0 * x / z, 360.0 + 780.0 * y / z]


def _make_dataset() -> vc.PlanarDataset:
    board = [(i * 0.04, j * 0.04, 0.0) for i in range(8) for j in range(6)]
    poses = [
        _make_pose(0.02 * i, -0.18 + 0.04 * i, -0.20 + 0.06 * i, -0.04 + 0.02 * i, 0.02 - 0.01 * i, 0.55 + 0.04 * i)
        for i in range(8)
    ]
    views: list[vc.PlanarView] = []
    for pose in poses:
        points_3d = []
        points_2d = []
        for point in board:
            point_cam = _transform_point(pose, list(point))
            uv = _project_point_pinhole(point_cam)
            if uv is None:
                continue
            points_3d.append(point)
            points_2d.append(tuple(uv))
        views.append(vc.PlanarView(observation=vc.Observation(points_3d=points_3d, points_2d=points_2d)))
    return vc.PlanarDataset(views=views)


class ScheimpflugIntrinsicsTest(unittest.TestCase):
    def test_public_bindings_run(self) -> None:
        result = vc.run_scheimpflug_intrinsics(
            _make_dataset(),
            vc.ScheimpflugIntrinsicsCalibrationConfig(
                fix_scheimpflug={"tilt_x": False, "tilt_y": False}
            ),
        )
        self.assertGreaterEqual(result.mean_reproj_error, 0.0)
        self.assertIsInstance(result.camera, vc.PinholeBrownConradyScheimpflugCamera)
        self.assertIsInstance(result.camera.sensor, vc.ScheimpflugSensor)

    def test_invalid_config_maps_to_set_config_error(self) -> None:
        with self.assertRaises(RuntimeError) as ctx:
            vc.run_scheimpflug_intrinsics(
                _make_dataset(),
                vc.ScheimpflugIntrinsicsCalibrationConfig(max_iters=0),
            )
        message = str(ctx.exception)
        self.assertIn("failed to set config", message)
        self.assertIn("max_iters must be positive", message)

    def test_invalid_input_maps_to_set_input_error(self) -> None:
        dataset = _make_dataset()
        dataset.views = dataset.views[:2]
        with self.assertRaises(RuntimeError) as ctx:
            vc.run_scheimpflug_intrinsics(
                dataset,
                vc.ScheimpflugIntrinsicsCalibrationConfig(),
            )
        message = str(ctx.exception)
        self.assertIn("failed to set input", message)
        self.assertIn("at least 3 views", message)

    def test_high_level_api_rejects_mapping_inputs(self) -> None:
        with self.assertRaises(TypeError) as cfg_ctx:
            vc.run_scheimpflug_intrinsics(
                _make_dataset(),
                {"max_iters": 50},
            )
        self.assertIn("config must be ScheimpflugIntrinsicsCalibrationConfig", str(cfg_ctx.exception))

        with self.assertRaises(TypeError) as input_ctx:
            vc.run_scheimpflug_intrinsics(
                {"views": []},
                vc.ScheimpflugIntrinsicsCalibrationConfig(),
            )
        self.assertIn("input must be PlanarDataset", str(input_ctx.exception))


if __name__ == "__main__":
    unittest.main()



from .models import (
    BrownConradyDistortion,
    LaserlineDataset,
    LaserlineDeviceCalibrationConfig,
    LaserlineEstimate,
    LaserlineEstimateParams,
    LaserlineDeviceInitConfig,
    LaserlineDeviceOptimizeConfig,
    LaserlineDeviceResult,
    LaserlineDeviceSolverConfig,
    LaserlinePlane,
    LaserlineStats,
    LaserlineView,
    Observation,
    PinholeBrownConradyCamera,
    PinholeBrownConradyScheimpflugCamera,
    PinholeIntrinsics,
    PlanarCalibrationConfig,
    PlanarCalibrationResult,
    PlanarDataset,
    PlanarView,
    Pose,
    ScheimpflugIntrinsicsCalibrationConfig,
    ScheimpflugSensor,
    ScheimpflugIntrinsicsResult,
    RigExtrinsicsCalibrationConfig,
    RigExtrinsicsDataset,
    RigExtrinsicsResult,
    RigExtrinsicsView,
    RigHandeyeBaConfig,
    RigHandeyeCalibrationConfig,
    RigHandeyeDataset,
    RigHandeyeInitConfig,
    RigHandeyeIntrinsicsConfig,
    RigHandeyeResult,
    RigHandeyeRigConfig,
    RigHandeyeSolverConfig,
    RigHandeyeView,
    SingleCamHandeyeCalibrationConfig,
    SingleCamHandeyeDataset,
    SingleCamHandeyeResult,
    SingleCamHandeyeView,
)
from .types import RobustLoss as _RobustLoss

__version__: str


def robust_none() -> _RobustLoss: ...
def robust_huber(scale: float) -> _RobustLoss: ...
def robust_cauchy(scale: float) -> _RobustLoss: ...
def robust_arctan(scale: float) -> _RobustLoss: ...


def run_planar_intrinsics(
    input: PlanarDataset,
    config: PlanarCalibrationConfig | None = None,
) -> PlanarCalibrationResult: ...


def run_scheimpflug_intrinsics(
    input: PlanarDataset,
    config: ScheimpflugIntrinsicsCalibrationConfig | None = None,
) -> ScheimpflugIntrinsicsResult: ...


def run_single_cam_handeye(
    input: SingleCamHandeyeDataset,
    config: SingleCamHandeyeCalibrationConfig | None = None,
) -> SingleCamHandeyeResult: ...


def run_rig_extrinsics(
    input: RigExtrinsicsDataset,
    config: RigExtrinsicsCalibrationConfig | None = None,
) -> RigExtrinsicsResult: ...


def run_rig_handeye(
    input: RigHandeyeDataset,
    config: RigHandeyeCalibrationConfig | None = None,
) -> RigHandeyeResult: ...


def run_laserline_device(
    input: LaserlineDataset,
    config: LaserlineDeviceCalibrationConfig | None = None,
) -> LaserlineDeviceResult: ...


__all__: list[str]



"""Multi-camera rig hand-eye calibration with synthetic data.

Python counterpart of:
`crates/vision-calibration/examples/rig_handeye_synthetic.rs`
"""

from __future__ import annotations

import vision_calibration as vc

from _common import (
    baseline_m,
    grid_points,
    make_pose,
    pose_compose,
    pose_inverse,
    project_point_pinhole,
    transform_point,
)


def project_view(
    camera_intrinsics: dict,
    camera_distortion: dict,
    cam_se3_target: dict,
    board_points: list[list[float]],
) -> vc.Observation | None:
    points_3d = []
    points_2d = []
    for p in board_points:
        p_cam = transform_point(cam_se3_target, p)
        if p_cam[2] <= 0.0:
            continue
        uv = project_point_pinhole(p_cam, camera_intrinsics, camera_distortion)
        if uv is None:
            continue
        if 0.0 <= uv[0] < 1280.0 and 0.0 <= uv[1] < 720.0:
            points_3d.append(p)
            points_2d.append(uv)
    if len(points_2d) < 4:
        return None
    return vc.Observation(points_3d=points_3d, points_2d=points_2d)


def make_input() -> vc.RigHandeyeDataset:
    k0 = {"fx": 800.0, "fy": 780.0, "cx": 640.0, "cy": 360.0, "skew": 0.0}
    k1 = {"fx": 810.0, "fy": 790.0, "cx": 635.0, "cy": 355.0, "skew": 0.0}
    dist = {"k1": 0.0, "k2": 0.0, "k3": 0.0, "p1": 0.0, "p2": 0.0, "iters": 8}

    cam0_se3_rig = make_pose(0.0, 0.0, 0.0, 0.00, 0.00, 0.00)  # T_C0_R
    cam1_se3_rig = make_pose(0.0, 0.0, 0.0, 0.12, 0.00, 0.00)  # T_C1_R
    handeye_gt = make_pose(0.05, -0.03, 0.10, 0.00, -0.05, 0.08)  # T_G_R
    target_in_base_gt = make_pose(0.0, 0.0, 0.0, 0.0, 0.0, 1.0)  # T_B_T

    board = grid_points(cols=6, rows=5, spacing_m=0.05)
    robot_poses = [
        make_pose(0.0, 0.0, 0.0, 0.00, 0.00, 0.00),
        make_pose(0.15, 0.0, 0.0, 0.05, 0.00, 0.00),
        make_pose(0.0, 0.15, 0.0, 0.00, 0.05, 0.00),
        make_pose(0.0, 0.0, 0.15, 0.00, 0.00, 0.05),
        make_pose(0.10, 0.10, 0.0, 0.03, -0.03, 0.00),
        make_pose(-0.10, 0.0, 0.10, -0.03, 0.03, 0.00),
        make_pose(0.08, -0.08, 0.05, 0.00, -0.03, 0.02),
    ]

    views: list[vc.RigHandeyeView] = []
    for robot_pose in robot_poses:
        # Eye-in-hand rig chain: T_R_T = (T_B_G * T_G_R)^-1 * T_B_T
        rig_se3_target = pose_compose(pose_inverse(pose_compose(robot_pose, handeye_gt)), target_in_base_gt)
        cam0_se3_target = pose_compose(cam0_se3_rig, rig_se3_target)
        cam1_se3_target = pose_compose(cam1_se3_rig, rig_se3_target)
        obs0 = project_view(k0, dist, cam0_se3_target, board)
        obs1 = project_view(k1, dist, cam1_se3_target, board)
        if obs0 is None and obs1 is None:
            continue
        views.append(
            vc.RigHandeyeView(
                cameras=[obs0, obs1],
                base_se3_gripper=vc.Pose(
                    rotation_xyzw=tuple(robot_pose["rotation"]),
                    translation_xyz=tuple(robot_pose["translation"]),
                ),
            )
        )

    return vc.RigHandeyeDataset(num_cameras=2, views=views)


def main() -> None:
    print("=== Rig Hand-Eye Calibration (Synthetic, Python) ===")
    payload = make_input()
    result = vc.run_rig_handeye(payload)
    base_mm = baseline_m(result.cam_se3_rig) * 1000.0

    print(f"Views: {len(payload.views)}, cameras: {payload.num_cameras}")
    print(f"Hand-eye mode: {result.handeye_mode}")
    print(f"Mean reprojection error: {result.mean_reproj_error:.6f} px")
    print(f"Per-camera reprojection: {result.per_cam_reproj_errors}")
    print(f"Estimated baseline: {base_mm:.3f} mm")
    if result.gripper_se3_rig is not None:
        t = result.gripper_se3_rig.translation_xyz
        print(f"Recovered gripper->rig translation: [{t[0]:.4f}, {t[1]:.4f}, {t[2]:.4f}] m")


if __name__ == "__main__":
    main()



"""Low-level serde payload contracts for :mod:`vision_calibration`.

These aliases/TypedDicts model the raw Rust JSON payload schema. Most users
should use dataclasses from :mod:`vision_calibration.models` instead.

Warning
-------
This module is compatibility-oriented low-level surface. It is not the
recommended high-level API for new code.
"""

from __future__ import annotations

from typing import Any, Literal, TypeAlias, TypedDict

JsonObject: TypeAlias = dict[str, Any]
JsonArray: TypeAlias = list[Any]

# Nested geometry/model payloads are serde dictionaries coming from Rust.
Transform: TypeAlias = JsonObject
CameraModel: TypeAlias = JsonObject
IntrinsicsPayload: TypeAlias = JsonObject
DistortionPayload: TypeAlias = JsonObject
ScheimpflugPayload: TypeAlias = JsonObject
LaserPlanePayload: TypeAlias = JsonObject

Se3Delta: TypeAlias = tuple[float, float, float, float, float, float] | list[float]

HandEyeMode: TypeAlias = Literal["EyeInHand", "EyeToHand"]
LaserlineResidualType: TypeAlias = Literal["PointToPlane", "LineDistNormalized"]
ScheimpflugFixMask: TypeAlias = JsonObject


class _RobustLossScale(TypedDict):
    scale: float


class RobustLossHuber(TypedDict):
    Huber: _RobustLossScale


class RobustLossCauchy(TypedDict):
    Cauchy: _RobustLossScale


class RobustLossArctan(TypedDict):
    Arctan: _RobustLossScale


# Serde forms accepted by Rust:
# - "None"
# - {"Huber": {"scale": float}}
# - {"Cauchy": {"scale": float}}
# - {"Arctan": {"scale": float}}
RobustLoss: TypeAlias = (
    Literal["None"]
    | RobustLossHuber
    | RobustLossCauchy
    | RobustLossArctan
)

# Input payload aliases (workflow datasets).
PlanarInput: TypeAlias = JsonObject
SingleCamHandeyeInput: TypeAlias = JsonObject
RigExtrinsicsInput: TypeAlias = JsonObject
RigHandeyeInput: TypeAlias = JsonObject
LaserlineDeviceInput: TypeAlias = JsonObject
ScheimpflugIntrinsicsInput: TypeAlias = JsonObject


class SolveReport(TypedDict):
    final_cost: float


class PlanarIntrinsicsParams(TypedDict):
    camera: CameraModel
    camera_se3_target: list[Transform]


class PlanarIntrinsicsExport(TypedDict):
    params: PlanarIntrinsicsParams
    report: SolveReport
    mean_reproj_error: float
    per_cam_reproj_errors: list[float]


class SingleCamHandeyeExport(TypedDict):
    camera: CameraModel
    handeye_mode: HandEyeMode
    gripper_se3_camera: Transform | None
    camera_se3_base: Transform | None
    base_se3_target: Transform | None
    gripper_se3_target: Transform | None
    robot_deltas: list[Se3Delta] | None
    mean_reproj_error: float
    per_cam_reproj_errors: list[float]


class RigExtrinsicsExport(TypedDict):
    cameras: list[CameraModel]
    cam_se3_rig: list[Transform]
    mean_reproj_error: float
    per_cam_reproj_errors: list[float]


class RigHandeyeExport(TypedDict):
    cameras: list[CameraModel]
    cam_se3_rig: list[Transform]
    handeye_mode: HandEyeMode
    gripper_se3_rig: Transform | None
    rig_se3_base: Transform | None
    base_se3_target: Transform | None
    gripper_se3_target: Transform | None
    robot_deltas: list[Se3Delta] | None
    mean_reproj_error: float
    per_cam_reproj_errors: list[float]


class LaserlineParams(TypedDict):
    intrinsics: IntrinsicsPayload
    distortion: DistortionPayload
    sensor: ScheimpflugPayload
    poses: list[Transform]
    plane: LaserPlanePayload


class LaserlineEstimate(TypedDict):
    params: LaserlineParams
    report: SolveReport


class LaserlineStats(TypedDict):
    mean_reproj_error: float
    mean_laser_error: float
    per_view_reproj_errors: list[float]
    per_view_laser_errors: list[float]


class LaserlineDeviceExport(TypedDict):
    estimate: LaserlineEstimate
    stats: LaserlineStats
    mean_reproj_error: float
    per_cam_reproj_errors: list[float]


class ScheimpflugIntrinsicsParams(TypedDict):
    camera: CameraModel
    camera_se3_target: list[Transform]


class ScheimpflugIntrinsicsExport(TypedDict):
    params: ScheimpflugIntrinsicsParams
    report: SolveReport
    mean_reproj_error: float
    per_cam_reproj_errors: list[float]


class PlanarIntrinsicsConfig(TypedDict, total=False):
    init_iterations: int
    fix_k3_in_init: bool
    fix_tangential_in_init: bool
    zero_skew: bool
    max_iters: int
    verbosity: int
    robust_loss: RobustLoss
    fix_intrinsics: JsonObject
    fix_distortion: JsonObject
    fix_poses: list[int]


class SingleCamHandeyeConfig(TypedDict, total=False):
    intrinsics_init_iterations: int
    fix_k3: bool
    fix_tangential: bool
    zero_skew: bool
    handeye_mode: HandEyeMode
    min_motion_angle_deg: float
    max_iters: int
    verbosity: int
    robust_loss: RobustLoss
    refine_robot_poses: bool
    robot_rot_sigma: float
    robot_trans_sigma: float


class RigExtrinsicsConfig(TypedDict, total=False):
    intrinsics_init_iterations: int
    fix_k3: bool
    fix_tangential: bool
    zero_skew: bool
    reference_camera_idx: int
    max_iters: int
    verbosity: int
    robust_loss: RobustLoss
    refine_intrinsics_in_rig_ba: bool
    fix_first_rig_pose: bool


class RigHandeyeIntrinsicsConfig(TypedDict, total=False):
    init_iterations: int
    fix_k3: bool
    fix_tangential: bool
    zero_skew: bool


class RigHandeyeRigConfig(TypedDict, total=False):
    reference_camera_idx: int
    refine_intrinsics_in_rig_ba: bool
    fix_first_rig_pose: bool


class RigHandeyeInitConfig(TypedDict, total=False):
    handeye_mode: HandEyeMode
    min_motion_angle_deg: float


class RigHandeyeSolverConfig(TypedDict, total=False):
    max_iters: int
    verbosity: int
    robust_loss: RobustLoss


class RigHandeyeBaConfig(TypedDict, total=False):
    refine_robot_poses: bool
    robot_rot_sigma: float
    robot_trans_sigma: float
    refine_cam_se3_rig_in_handeye_ba: bool


class RigHandeyeConfig(TypedDict, total=False):
    intrinsics: RigHandeyeIntrinsicsConfig
    rig: RigHandeyeRigConfig
    handeye_init: RigHandeyeInitConfig
    solver: RigHandeyeSolverConfig
    handeye_ba: RigHandeyeBaConfig


class LaserlineDeviceInitConfig(TypedDict, total=False):
    iterations: int
    fix_k3: bool
    fix_tangential: bool
    zero_skew: bool
    sensor_init: JsonObject


class LaserlineDeviceSolverConfig(TypedDict, total=False):
    max_iters: int
    verbosity: int


class LaserlineDeviceOptimizeConfig(TypedDict, total=False):
    calib_loss: RobustLoss
    laser_loss: RobustLoss
    calib_weight: float
    laser_weight: float
    fix_intrinsics: bool
    fix_distortion: bool
    fix_k3: bool
    fix_sensor: bool
    fix_poses: list[int]
    fix_plane: bool
    laser_residual_type: LaserlineResidualType


class LaserlineDeviceConfig(TypedDict, total=False):
    init: LaserlineDeviceInitConfig
    solver: LaserlineDeviceSolverConfig
    optimize: LaserlineDeviceOptimizeConfig


class ScheimpflugIntrinsicsConfig(TypedDict, total=False):
    init_iterations: int
    fix_k3_in_init: bool
    zero_skew: bool
    max_iters: int
    verbosity: int
    robust_loss: RobustLoss
    fix_intrinsics: JsonObject
    fix_distortion: JsonObject
    fix_scheimpflug: ScheimpflugFixMask
    fix_first_pose: bool



from __future__ import annotations

import unittest

import vision_calibration as vc
import vision_calibration.types as vc_types


class TypedModelParsingTest(unittest.TestCase):
    def test_types_module_is_low_level_and_not_reexported_top_level(self) -> None:
        self.assertFalse(hasattr(vc, "RobustLoss"))
        self.assertFalse(hasattr(vc, "HandEyeMode"))
        self.assertTrue(hasattr(vc_types, "RobustLoss"))

    def test_pinhole_camera_from_runtime_payload(self) -> None:
        payload = {
            "proj": "Pinhole",
            "k": {
                "fx": 800.0,
                "fy": 780.0,
                "cx": 640.0,
                "cy": 360.0,
                "skew": 0.0,
            },
            "dist": {
                "k1": 0.1,
                "k2": -0.01,
                "k3": 0.001,
                "p1": 0.0,
                "p2": 0.0,
                "iters": 8,
            },
        }

        cam = vc.PinholeBrownConradyCamera.from_payload(payload)
        self.assertAlmostEqual(cam.intrinsics.fx, 800.0)
        self.assertAlmostEqual(cam.intrinsics.fy, 780.0)
        self.assertAlmostEqual(cam.distortion.k1, 0.1)
        self.assertEqual(cam.distortion.iters, 8)

    def test_scheimpflug_sensor_parses_legacy_aliases(self) -> None:
        sensor = vc.ScheimpflugSensor.from_payload({"tau_x": 0.01, "tau_y": -0.02})
        self.assertAlmostEqual(sensor.tilt_x, 0.01)
        self.assertAlmostEqual(sensor.tilt_y, -0.02)

    def test_scheimpflug_camera_from_camera_params_payload(self) -> None:
        payload = {
            "projection": {"type": "pinhole"},
            "intrinsics": {
                "type": "fx_fy_cx_cy_skew",
                "fx": 800.0,
                "fy": 780.0,
                "cx": 640.0,
                "cy": 360.0,
                "skew": 0.0,
            },
            "distortion": {
                "type": "brown_conrady5",
                "k1": 0.01,
                "k2": -0.002,
                "k3": 0.0,
                "p1": 0.0,
                "p2": 0.0,
                "iters": 8,
            },
            "sensor": {
                "type": "scheimpflug",
                "tilt_x": 0.015,
                "tilt_y": -0.012,
            },
        }

        cam = vc.PinholeBrownConradyScheimpflugCamera.from_payload(payload)
        self.assertAlmostEqual(cam.sensor.tilt_x, 0.015)
        self.assertAlmostEqual(cam.sensor.tilt_y, -0.012)


if __name__ == "__main__":
    unittest.main()



"""Shared helpers for vision-calibration Python examples.

These examples intentionally use only Python stdlib so they run in a minimal
virtual environment where `vision_calibration` is installed.
"""

from __future__ import annotations

import math
from pathlib import Path
from typing import Iterable


def repo_root() -> Path:
    return Path(__file__).resolve().parents[3]


def norm3(v: Iterable[float]) -> float:
    x, y, z = v
    return math.sqrt(x * x + y * y + z * z)


def add3(a: list[float], b: list[float]) -> list[float]:
    return [a[0] + b[0], a[1] + b[1], a[2] + b[2]]


def sub3(a: list[float], b: list[float]) -> list[float]:
    return [a[0] - b[0], a[1] - b[1], a[2] - b[2]]


def scale3(v: list[float], s: float) -> list[float]:
    return [v[0] * s, v[1] * s, v[2] * s]


def dot3(a: list[float], b: list[float]) -> float:
    return a[0] * b[0] + a[1] * b[1] + a[2] * b[2]


def normalize3(v: list[float]) -> list[float]:
    n = norm3(v)
    if n == 0.0:
        raise ValueError("cannot normalize zero-length vector")
    return [v[0] / n, v[1] / n, v[2] / n]


def mat3_mul_vec(m: list[list[float]], v: list[float]) -> list[float]:
    return [
        m[0][0] * v[0] + m[0][1] * v[1] + m[0][2] * v[2],
        m[1][0] * v[0] + m[1][1] * v[1] + m[1][2] * v[2],
        m[2][0] * v[0] + m[2][1] * v[1] + m[2][2] * v[2],
    ]


def mat3_mul(a: list[list[float]], b: list[list[float]]) -> list[list[float]]:
    return [
        [
            a[0][0] * b[0][0] + a[0][1] * b[1][0] + a[0][2] * b[2][0],
            a[0][0] * b[0][1] + a[0][1] * b[1][1] + a[0][2] * b[2][1],
            a[0][0] * b[0][2] + a[0][1] * b[1][2] + a[0][2] * b[2][2],
        ],
        [
            a[1][0] * b[0][0] + a[1][1] * b[1][0] + a[1][2] * b[2][0],
            a[1][0] * b[0][1] + a[1][1] * b[1][1] + a[1][2] * b[2][1],
            a[1][0] * b[0][2] + a[1][1] * b[1][2] + a[1][2] * b[2][2],
        ],
        [
            a[2][0] * b[0][0] + a[2][1] * b[1][0] + a[2][2] * b[2][0],
            a[2][0] * b[0][1] + a[2][1] * b[1][1] + a[2][2] * b[2][1],
            a[2][0] * b[0][2] + a[2][1] * b[1][2] + a[2][2] * b[2][2],
        ],
    ]


def mat3_transpose(m: list[list[float]]) -> list[list[float]]:
    return [
        [m[0][0], m[1][0], m[2][0]],
        [m[0][1], m[1][1], m[2][1]],
        [m[0][2], m[1][2], m[2][2]],
    ]


def rotation_from_euler_xyz(ax: float, ay: float, az: float) -> list[list[float]]:
    sx, cx = math.sin(ax), math.cos(ax)
    sy, cy = math.sin(ay), math.cos(ay)
    sz, cz = math.sin(az), math.cos(az)
    return [
        [cz * cy, cz * sy * sx - sz * cx, cz * sy * cx + sz * sx],
        [sz * cy, sz * sy * sx + cz * cx, sz * sy * cx - cz * sx],
        [-sy, cy * sx, cy * cx],
    ]


def quat_xyzw_from_rot(r: list[list[float]]) -> list[float]:
    trace = r[0][0] + r[1][1] + r[2][2]
    if trace > 0.0:
        s = math.sqrt(trace + 1.0) * 2.0
        w = 0.25 * s
        x = (r[2][1] - r[1][2]) / s
        y = (r[0][2] - r[2][0]) / s
        z = (r[1][0] - r[0][1]) / s
    elif r[0][0] > r[1][1] and r[0][0] > r[2][2]:
        s = math.sqrt(1.0 + r[0][0] - r[1][1] - r[2][2]) * 2.0
        w = (r[2][1] - r[1][2]) / s
        x = 0.25 * s
        y = (r[0][1] + r[1][0]) / s
        z = (r[0][2] + r[2][0]) / s
    elif r[1][1] > r[2][2]:
        s = math.sqrt(1.0 + r[1][1] - r[0][0] - r[2][2]) * 2.0
        w = (r[0][2] - r[2][0]) / s
        x = (r[0][1] + r[1][0]) / s
        y = 0.25 * s
        z = (r[1][2] + r[2][1]) / s
    else:
        s = math.sqrt(1.0 + r[2][2] - r[0][0] - r[1][1]) * 2.0
        w = (r[1][0] - r[0][1]) / s
        x = (r[0][2] + r[2][0]) / s
        y = (r[1][2] + r[2][1]) / s
        z = 0.25 * s
    return [x, y, z, w]


def make_pose(ax: float, ay: float, az: float, tx: float, ty: float, tz: float) -> dict:
    r = rotation_from_euler_xyz(ax, ay, az)
    q = quat_xyzw_from_rot(r)
    return {"rotation": q, "translation": [tx, ty, tz]}


def pose_to_rt(pose: dict) -> tuple[list[list[float]], list[float]]:
    x, y, z, w = pose["rotation"]
    tx, ty, tz = pose["translation"]
    xx = x * x
    yy = y * y
    zz = z * z
    xy = x * y
    xz = x * z
    yz = y * z
    wx = w * x
    wy = w * y
    wz = w * z
    r = [
        [1.0 - 2.0 * (yy + zz), 2.0 * (xy - wz), 2.0 * (xz + wy)],
        [2.0 * (xy + wz), 1.0 - 2.0 * (xx + zz), 2.0 * (yz - wx)],
        [2.0 * (xz - wy), 2.0 * (yz + wx), 1.0 - 2.0 * (xx + yy)],
    ]
    return r, [tx, ty, tz]


def pose_compose(a: dict, b: dict) -> dict:
    ra, ta = pose_to_rt(a)
    rb, tb = pose_to_rt(b)
    r = mat3_mul(ra, rb)
    t = add3(mat3_mul_vec(ra, tb), ta)
    return {"rotation": quat_xyzw_from_rot(r), "translation": t}


def pose_inverse(pose: dict) -> dict:
    r, t = pose_to_rt(pose)
    rt = mat3_transpose(r)
    tinv = scale3(mat3_mul_vec(rt, t), -1.0)
    return {"rotation": quat_xyzw_from_rot(rt), "translation": tinv}


def transform_point(pose: dict, p: list[float]) -> list[float]:
    r, t = pose_to_rt(pose)
    return add3(mat3_mul_vec(r, p), t)


def project_point_pinhole(
    p_cam: list[float],
    intrinsics: dict,
    distortion: dict | None = None,
) -> list[float] | None:
    x, y, z = p_cam
    if z <= 0.0:
        return None
    xn = x / z
    yn = y / z
    if distortion is not None:
        k1 = float(distortion["k1"])
        k2 = float(distortion["k2"])
        k3 = float(distortion["k3"])
        p1 = float(distortion["p1"])
        p2 = float(distortion["p2"])
        r2 = xn * xn + yn * yn
        r4 = r2 * r2
        r6 = r4 * r2
        radial = 1.0 + k1 * r2 + k2 * r4 + k3 * r6
        x_tan = 2.0 * p1 * xn * yn + p2 * (r2 + 2.0 * xn * xn)
        y_tan = p1 * (r2 + 2.0 * yn * yn) + 2.0 * p2 * xn * yn
        xd = xn * radial + x_tan
        yd = yn * radial + y_tan
    else:
        xd = xn
        yd = yn

    fx = float(intrinsics["fx"])
    fy = float(intrinsics["fy"])
    cx = float(intrinsics["cx"])
    cy = float(intrinsics["cy"])
    skew = float(intrinsics.get("skew", 0.0))
    u = fx * xd + skew * yd + cx
    v = fy * yd + cy
    return [u, v]


def grid_points(cols: int, rows: int, spacing_m: float) -> list[list[float]]:
    points: list[list[float]] = []
    for i in range(cols):
        for j in range(rows):
            points.append([i * spacing_m, j * spacing_m, 0.0])
    return points


def baseline_m(cam_se3_rig: list[object]) -> float:
    if len(cam_se3_rig) < 2:
        return 0.0
    def _translation(pose: object) -> list[float]:
        if isinstance(pose, dict):
            return list(pose["translation"])
        # vision_calibration.models.Pose
        return list(getattr(pose, "translation_xyz"))

    t0 = _translation(cam_se3_rig[0])
    t1 = _translation(cam_se3_rig[1])
    return norm3(sub3(t1, t0))
